# distutils: language = c++

"""These functions provide increased speed in handling the signal-binning functions necessary for supporting fivec analysis using HiFive."""

import cython
cimport numpy as np
import numpy

ctypedef np.float32_t DTYPE_t
ctypedef np.float64_t DTYPE_64_t
ctypedef np.int32_t DTYPE_int_t
ctypedef np.int64_t DTYPE_int64_t
ctypedef np.uint32_t DTYPE_uint_t
ctypedef np.int8_t DTYPE_int8_t
cdef double Inf = numpy.inf

cdef extern from "math.h":
    double exp(double x) nogil
    double log(double x) nogil
    double log10(double x) nogil
    double sqrt(double x) nogil
    double pow(double x, double x) nogil
    double abs(double x) nogil
    double round(double x) nogil
    double floor(double x) nogil
    double ceil(double x) nogil


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def find_cis_compact_observed(
        np.ndarray[DTYPE_int_t, ndim=2] data,
        np.ndarray[DTYPE_int64_t, ndim=1] indices,
        np.ndarray[DTYPE_int_t, ndim=1] mapping not None,
        np.ndarray[DTYPE_t, ndim=3] signal not None):
    cdef long long int frag1, frag2, i, map1, map2
    cdef long long int num_frags = mapping.shape[0]
    with nogil:
        for frag1 in range(num_frags - 1):
            map1 = mapping[frag1]
            if map1 == 0:
                continue
            if map1 < 0:
                map1 = -1 - map1
            else:
                map1 = map1 - 1
            for i in range(indices[frag1], indices[frag1 + 1]):
                frag2 = data[i, 1]
                if frag2 >= num_frags:
                    continue
                map2 = mapping[frag2]
                if map2 == 0:
                    continue
                if map2 < 0:
                    signal[map1, -1 - map2, 0] += data[i, 2]
                else:
                    signal[map2 - 1, map1, 0] += data[i, 2]
    return None


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def find_cis_upper_observed(
        np.ndarray[DTYPE_int_t, ndim=2] data,
        np.ndarray[DTYPE_int64_t, ndim=1] indices,
        np.ndarray[DTYPE_int_t, ndim=1] mapping not None,
        np.ndarray[DTYPE_t, ndim=2] signal not None):
    cdef long long int frag1, frag2, i, index, map1, map2
    cdef long long int num_frags = mapping.shape[0]
    cdef long long int num_bins = int(0.5 + pow(0.25 + 2 * signal.shape[0], 0.5))
    with nogil:
        for frag1 in range(num_frags - 1):
            map1 = mapping[frag1]
            if map1 == -1:
                continue
            index = map1 * (num_bins - 1) - map1 * (map1 + 1) / 2 - 1
            for i in range(indices[frag1], indices[frag1 + 1]):
                frag2 = data[i, 1]
                if frag2 >= num_frags:
                    continue
                map2 = mapping[frag2]
                if map2 == -1 or map2 == map1:
                    continue
                signal[index + map2, 0] += data[i, 2]
    return None


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def find_cis_compact_expected(
        np.ndarray[DTYPE_int_t, ndim=1] mapping not None,
        np.ndarray[DTYPE_t, ndim=1] corrections,
        np.ndarray[DTYPE_t, ndim=1] binning_corrections,
        np.ndarray[DTYPE_int_t, ndim=1] correction_indices,
        np.ndarray[DTYPE_int_t, ndim=1] binning_num_bins,
        np.ndarray[DTYPE_int_t, ndim=2] frag_indices,
        np.ndarray[DTYPE_int_t, ndim=1] mids,
        np.ndarray[DTYPE_int_t, ndim=1] strands,
        np.ndarray[DTYPE_t, ndim=3] signal not None,
        double gamma,
        double region_mean,
        int startfrag):
    cdef long long int frag1, frag2, map1, map2, strand1, bin1, bin2, index, num_parameters
    cdef double distance, value
    cdef long long int num_frags = mapping.shape[0]
    cdef long long int num_bins = int(0.5 + pow(0.25 + 2 * signal.shape[0], 0.5))
    if not frag_indices is None:
        num_parameters = frag_indices.shape[1]
    else:
        num_parameters = 0
    with nogil:
        for frag1 in range(num_frags - 1):
            map1 = mapping[frag1]
            if map1 == 0:
                continue
            if map1 < 0:
                map1 = -1 - map1
            else:
                map1 = map1 - 1
            strand1 = strands[frag1]
            for frag2 in range(frag1 + 1, num_frags):
                map2 = mapping[frag2]
                if strands[frag2] == strand1 or map2 == 0:
                    continue
                 # give starting expected value
                value = 0.0
                # if finding frag, enrichment, or expected, and using express or probability bias correction, correct for frag
                if not corrections is None:
                    value += corrections[frag1] + corrections[frag2]
                # if finding frag, enrichment, or expected, and using binning bias correction, correct for frag
                if not binning_corrections is None:
                    for j in range(num_parameters):
                        bin1 = min(frag_indices[frag1 + startfrag, j], frag_indices[frag2 + startfrag, j]) 
                        bin2 = max(frag_indices[frag1 + startfrag, j], frag_indices[frag2 + startfrag, j]) 
                        index = bin1 * (binning_num_bins[j] - 1) - bin1 * (bin1 - 1) / 2 + bin2 + correction_indices[j]
                        value += binning_corrections[index]
                # if finding distance, enrichment, or expected, correct for distance
                if gamma != 0.0:
                    distance = log(<double>(mids[frag2] - mids[frag1]))
                    value += region_mean - distance * gamma
                if strand1 == 0:
                    signal[map1, -1 - map2, 1] += exp(value)
                else:
                    signal[map2 - 1, map1, 1] += exp(value)
    return None


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def find_cis_upper_expected(
        np.ndarray[DTYPE_int_t, ndim=1] mapping not None,
        np.ndarray[DTYPE_t, ndim=1] corrections,
        np.ndarray[DTYPE_t, ndim=1] binning_corrections,
        np.ndarray[DTYPE_int_t, ndim=1] correction_indices,
        np.ndarray[DTYPE_int_t, ndim=1] binning_num_bins,
        np.ndarray[DTYPE_int_t, ndim=2] frag_indices,
        np.ndarray[DTYPE_int_t, ndim=1] mids not None,
        np.ndarray[DTYPE_int_t, ndim=1] strands not None,
        np.ndarray[DTYPE_t, ndim=2] signal not None,
        double gamma,
        double region_mean,
        int startfrag):
    cdef long long int frag1, frag2, index, map1, map2, strand1, bin1, bin2, index2, num_parameters
    cdef double distance, value
    cdef long long int num_frags = mapping.shape[0]
    cdef long long int num_bins = int(0.5 + pow(0.25 + 2 * signal.shape[0], 0.5))
    if not frag_indices is None:
        num_parameters = frag_indices.shape[1]
    else:
        num_parameters = 0
    with nogil:
        for frag1 in range(num_frags - 1):
            map1 = mapping[frag1]
            if map1 == -1:
                continue
            strand1 = strands[frag1]
            index = map1 * (num_bins - 1) - map1 * (map1 + 1) / 2 - 1
            for frag2 in range(frag1 + 1, num_frags):
                map2 = mapping[frag2]
                if strands[frag2] == strand1 or map2 == -1 or map2 == map1:
                    continue
                 # give starting expected value
                value = 0.0
                # if finding frag, enrichment, or expected, and using express or probability bias correction, correct for frag
                if not corrections is None:
                    value += corrections[frag1] + corrections[frag2]
                # if finding frag, enrichment, or expected, and using binning bias correction, correct for frag
                if not binning_corrections is None:
                    for j in range(num_parameters):
                        bin1 = min(frag_indices[frag1 + startfrag, j], frag_indices[frag2 + startfrag, j]) 
                        bin2 = max(frag_indices[frag1 + startfrag, j], frag_indices[frag2 + startfrag, j]) 
                        index2 = bin1 * (binning_num_bins[j] - 1) - bin1 * (bin1 - 1) / 2 + bin2 + correction_indices[j]
                        value += binning_corrections[index2]
                # if finding distance, enrichment, or expected, correct for distance
                if gamma != 0.0:
                    distance = log(<double>(mids[frag2] - mids[frag1]))
                    value += region_mean - distance * gamma
                signal[index + map2, 1] += exp(value)
    return None


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def bin_upper_to_upper(
        np.ndarray[DTYPE_t, ndim=2] binned not None,
        np.ndarray[DTYPE_t, ndim=2] unbinned not None,
        np.ndarray[DTYPE_int_t, ndim=1] mapping not None,
        int num_bins):
    cdef long long int i, j, index, index2
    cdef long long int num_fends = mapping.shape[0]
    with nogil:
        for i in range(num_fends - 1):
            if mapping[i] == -1:
                continue
            index = i * num_fends - i * (i + 1) / 2 - i - 1
            index2 = mapping[i] * num_bins - mapping[i] * (mapping[i] + 1) / 2 - mapping[i] - 1
            for j in range(i + 1, num_fends):
                if mapping[j] == -1 or mapping[j] == mapping[i]:
                    continue
                else:
                    binned[index2 + mapping[j], 0] += unbinned[index + j, 0]
                    binned[index2 + mapping[j], 1] += unbinned[index + j, 1]
    return None


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def dynamically_bin_upper_from_upper(
        np.ndarray[DTYPE_t, ndim=2] unbinned not None,
        np.ndarray[DTYPE_int_t, ndim=1] ub_mids not None,
        np.ndarray[DTYPE_t, ndim=2] binned not None,
        np.ndarray[DTYPE_int_t, ndim=2] bounds not None,
        np.ndarray[DTYPE_int_t, ndim=1] b_mids not None,
        int minobservations,
        int maxsearch,
        int removefailed):
    cdef long long int x, y, i, lX, lX_dist, uX, uX_dist, lY, lY_dist, uY, uY_dist, min_dist, index, index2
    cdef long long int num_bins = bounds.shape[0]
    cdef long long int num_fends = ub_mids.shape[0]
    with nogil:
        for x in range(num_bins - 1):
            index = x * num_bins - x * (x + 1) / 2 - x - 1
            for y in range(x + 1, num_bins):
                # if bin already meets our criteria, skip
                if binned[index + y, 0] >= minobservations:
                    continue
                # otherwise, set boarding unbinned positions according to bounds
                lX = bounds[x, 0]
                uX = bounds[x, 1] - 1
                lY = bounds[y, 0]
                uY = bounds[y, 1] - 1
                # find distance in each direction
                if lX > 0:
                    lX_dist = b_mids[x] - ub_mids[lX - 1]
                else:
                    lX_dist = 1000000000
                if uX < num_fends - 1:
                    uX_dist = ub_mids[uX + 1] - b_mids[x]
                else:
                    uX_dist = 1000000000
                if lY > 0:
                    lY_dist = b_mids[y] - ub_mids[lY - 1]
                else:
                    lY_dist = 1000000000
                if uY < num_fends - 1:
                    uY_dist = ub_mids[uY + 1] - b_mids[y]
                else:
                    uY_dist = 1000000000
                # determine min distance
                min_dist = min(min(lX_dist, uX_dist), min(lY_dist, uY_dist))
                # keep searching while less than maxsearch and minobservations
                while (maxsearch == 0 or min_dist < maxsearch) and binned[index + y, 0] < minobservations:
                    # find min dist, update distance, and add new row or col of observations
                    if min_dist == lX_dist:
                        lX -= 1
                        if lX > 0:
                            lX_dist = b_mids[x] - ub_mids[lX - 1]
                        else:
                            lX_dist = 1000000000
                        index2 = lX * num_fends - lX * (lX + 1) / 2 - lX - 1
                        for i in range(max(lX + 1, lY), uY + 1):
                            binned[index + y, 0] += unbinned[index2 + i, 0]
                            binned[index + y, 1] += unbinned[index2 + i, 1]
                    if min_dist == uX_dist:
                        uX += 1
                        if uX < num_fends - 1:
                            uX_dist = ub_mids[uX + 1] - b_mids[x]
                        else:
                            uX_dist = 1000000000
                        index2 = uX * num_fends - uX * (uX + 1) / 2 - uX - 1
                        for i in range(max(uX + 1, lY), uY + 1):
                            binned[index + y, 0] += unbinned[index2 + i, 0]
                            binned[index + y, 1] += unbinned[index2 + i, 1]
                    if min_dist == lY_dist:
                        lY -= 1
                        if lY > 0:
                            lY_dist = b_mids[y] - ub_mids[lY - 1]
                        else:
                            lY_dist = 1000000000
                        for i in range(lX, min(uX + 1, lY)):
                            index2 = i * num_fends - i * (i + 1) / 2 - i - 1
                            binned[index + y, 0] += unbinned[index2 + lY, 0]
                            binned[index + y, 1] += unbinned[index2 + lY, 1]
                    if min_dist == uY_dist:
                        uY += 1
                        if uY < num_fends - 1:
                            uY_dist = ub_mids[uY + 1] - b_mids[y]
                        else:
                            uY_dist = 1000000000
                        for i in range(lX, min(uX + 1, uY)):
                            index2 = i * num_fends - i * (i + 1) / 2 - i - 1
                            binned[index + y, 0] += unbinned[index2 + uY, 0]
                            binned[index + y, 1] += unbinned[index2 + uY, 1]
                    # determine min distance
                    min_dist = min(min(lX_dist, uX_dist), min(lY_dist, uY_dist))
                if binned[index + y, 0] < minobservations and removefailed == 1:
                    binned[index + y, 0] = 0
                    binned[index + y, 1] = 0
    return None


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def find_trans_observed(
        np.ndarray[DTYPE_int_t, ndim=2] data,
        np.ndarray[DTYPE_int64_t, ndim=1] indices,
        np.ndarray[DTYPE_int_t, ndim=1] mapping1 not None,
        np.ndarray[DTYPE_int_t, ndim=1] mapping2 not None,
        np.ndarray[DTYPE_t, ndim=3] signal not None):
    cdef long long int frag1, frag2, i, map1, map2
    cdef long long int num_frags1 = mapping1.shape[0]
    cdef long long int num_frags2 = mapping2.shape[0]
    cdef long long int num_data = data.shape[0]
    with nogil:
        for frag1 in range(num_frags1):
            map1 = mapping1[frag1]
            if map1 == -1:
                continue
            for i in range(indices[frag1], indices[frag1 + 1]):
                frag2 = data[i, 1]
                if frag2 < 0 or frag2 >= num_frags2:
                    continue
                map2 = mapping2[frag2]
                if map2 == -1:
                    continue
                signal[map1, map2, 0] += data[i, 2]
        """
        for i in range(num_data):
            frag1 = data[i, 0]
            frag2 = data[i, 1]
            if frag2 < 0 or frag2 >= num_frags2:
                continue
            map1 = mapping1[frag1]
            map2 = mapping2[frag2]
            if map1 == -1 or map2 == -1:
                continue
            signal[map1, map2, 0] += data[i, 2]
        """
    return None


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def find_trans_expected(
        np.ndarray[DTYPE_int_t, ndim=1] mapping1 not None,
        np.ndarray[DTYPE_int_t, ndim=1] mapping2 not None,
        np.ndarray[DTYPE_t, ndim=1] corrections1,
        np.ndarray[DTYPE_t, ndim=1] corrections2,
        np.ndarray[DTYPE_t, ndim=1] binning_corrections,
        np.ndarray[DTYPE_int_t, ndim=1] correction_indices,
        np.ndarray[DTYPE_int_t, ndim=1] binning_num_bins,
        np.ndarray[DTYPE_int_t, ndim=2] frag_indices,
        np.ndarray[DTYPE_int_t, ndim=1] strands1,
        np.ndarray[DTYPE_int_t, ndim=1] strands2,
        np.ndarray[DTYPE_t, ndim=3] signal not None,
        double trans_mean,
        int startfrag1,
        int startfrag2):
    cdef long long int frag1, frag2, map1, map2, strand1, bin1, bin2, index, num_parameters
    cdef double distance, value
    cdef long long int num_frags1 = mapping1.shape[0]
    cdef long long int num_frags2 = mapping2.shape[0]
    if not frag_indices is None:
        num_parameters = frag_indices.shape[1]
    else:
        num_parameters = 0
    with nogil:
        for frag1 in range(num_frags1):
            map1 = mapping1[frag1]
            if map1 == -1:
                continue
            strand1 = strands1[frag1]
            for frag2 in range(num_frags2):
                map2 = mapping2[frag2]
                if strands2[frag2] == strand1 or map2 == -1:
                    continue
                 # give starting expected value
                value = trans_mean
                # if finding frag, enrichment, or expected, and using express or probability bias correction, correct for frag
                if not corrections1 is None:
                    value += corrections1[frag1] + corrections2[frag2]
                # if finding frag, enrichment, or expected, and using binning bias correction, correct for frag
                if not binning_corrections is None:
                    for j in range(num_parameters):
                        bin1 = min(frag_indices[frag1 + startfrag1, j], frag_indices[frag2 + startfrag2, j]) 
                        bin2 = max(frag_indices[frag1 + startfrag1, j], frag_indices[frag2 + startfrag2, j]) 
                        index = bin1 * (binning_num_bins[j] - 1) - bin1 * (bin1 - 1) / 2 + bin2 + correction_indices[j]
                        value += binning_corrections[index]
                signal[map1, map2, 1] += exp(value)
    return None


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def dynamically_bin_trans(
        np.ndarray[DTYPE_t, ndim=3] unbinned not None,
        np.ndarray[DTYPE_int_t, ndim=1] mids1 not None,
        np.ndarray[DTYPE_int_t, ndim=1] mids2 not None,
        np.ndarray[DTYPE_t, ndim=3] binned not None,
        np.ndarray[DTYPE_int_t, ndim=2] bounds1 not None,
        np.ndarray[DTYPE_int_t, ndim=2] bounds2 not None,
        np.ndarray[DTYPE_int_t, ndim=1] b_mids1 not None,
        np.ndarray[DTYPE_int_t, ndim=1] b_mids2 not None,
        int minobservations,
        int maxsearch,
        int removefailed):
    cdef long long int x, y, i, lX, lX_dist, uX, uX_dist, lY, lY_dist, uY, uY_dist, min_dist
    cdef long long int num_bins1 = bounds1.shape[0]
    cdef long long int num_bins2 = bounds2.shape[0]
    cdef long long int num_frags1 = mids1.shape[0]
    cdef long long int num_frags2 = mids2.shape[0]
    with nogil:
        for x in range(num_bins1):
            for y in range(num_bins2):
                # if bin already meets our criteria, skip
                if binned[x, y, 0] >= minobservations:
                    continue
                # otherwise, set boarding unbinned positions according to bounds
                lX = bounds1[x, 0]
                uX = bounds1[x, 1] - 1
                lY = bounds2[y, 0]
                uY = bounds2[y, 1] - 1
                # find distance in each direction
                if lX > 0:
                    lX_dist = b_mids1[x] - mids1[lX - 1]
                else:
                    lX_dist = 1000000000
                if uX < num_frags1 - 1:
                    uX_dist = mids1[uX + 1] - b_mids1[x]
                else:
                    uX_dist = 1000000000
                if lY > 0:
                    lY_dist = b_mids2[y] - mids2[lY - 1]
                else:
                    lY_dist = 1000000000
                if uY < num_frags2 - 1:
                    uY_dist = mids2[uY + 1] - b_mids2[y]
                else:
                    uY_dist = 1000000000
                # determine min distance
                min_dist = min(min(lX_dist, uX_dist), min(lY_dist, uY_dist))
                # keep searching while less than maxsearch and minobservations
                while (maxsearch == 0 or min_dist < maxsearch) and binned[x, y, 0] < minobservations:
                    # find min dist, update distance, and add new row or col of observations
                    if min_dist == lX_dist:
                        lX -= 1
                        if lX > 0:
                            lX_dist = b_mids1[x] - mids1[lX - 1]
                        else:
                            lX_dist = 1000000000
                        for i in range(lY, uY + 1):
                            binned[x, y, 0] += unbinned[lX, i, 0]
                            binned[x, y, 1] += unbinned[lX, i, 1]
                    if min_dist == uX_dist:
                        uX += 1
                        if uX < num_frags1 - 1:
                            uX_dist = mids1[uX + 1] - b_mids1[x]
                        else:
                            uX_dist = 1000000000
                        for i in range(lY, uY + 1):
                            binned[x, y, 0] += unbinned[uX, i, 0]
                            binned[x, y, 1] += unbinned[uX, i, 1]
                    if min_dist == lY_dist:
                        lY -= 1
                        if lY > 0:
                            lY_dist = b_mids2[y] - mids2[lY - 1]
                        else:
                            lY_dist = 1000000000
                        for i in range(lX, uX + 1):
                            binned[x, y, 0] += unbinned[i, lY, 0]
                            binned[x, y, 1] += unbinned[i, lY, 1]
                    if min_dist == uY_dist:
                        uY += 1
                        if uY < num_frags2 - 1:
                            uY_dist = mids2[uY + 1] - b_mids2[y]
                        else:
                            uY_dist = 1000000000
                        for i in range(lX, uX + 1):
                            binned[x, y, 0] += unbinned[i, uY, 0]
                            binned[x, y, 1] += unbinned[i, uY, 1]
                    # determine min distance
                    min_dist = min(min(lX_dist, uX_dist), min(lY_dist, uY_dist))
                if binned[x, y, 0] < minobservations and removefailed == 1:
                    binned[x, y, 0] = 0
                    binned[x, y, 1] = 0
    return None


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def binning_bin_observed(
        np.ndarray[DTYPE_int_t, ndim=2] data,
        np.ndarray[DTYPE_int_t, ndim=2] trans_data,
        np.ndarray[DTYPE_int_t, ndim=1] regions,
        np.ndarray[DTYPE_int_t, ndim=1] mids,
        np.ndarray[DTYPE_int64_t, ndim=1] counts,
        np.ndarray[DTYPE_64_t, ndim=2] sums,
        np.ndarray[DTYPE_int_t, ndim=2] all_indices,
        np.ndarray[DTYPE_int_t, ndim=1] num_bins,
        np.ndarray[DTYPE_int_t, ndim=1] bin_divs,
        np.ndarray[DTYPE_t, ndim=1] region_means,
        np.ndarray[DTYPE_t, ndim=1] corrections,
        int mindistance,
        int maxdistance,
        float gamma,
        float trans_mean):
    cdef long long int i, k, distance, frag1, frag2, index, bin1, bin2, prev_frag, num_data, num_trans
    cdef double log_dist, signal, log_count
    cdef long long int num_features = all_indices.shape[1]
    if not data is None:
        num_data = data.shape[0]
    else:
        num_data = 0
    if not trans_data is None:
        num_trans = trans_data.shape[0]
    else:
        num_trans = 0
    with nogil:
        prev_frag = -1
        for i in range(num_data):
            frag1 = data[i, 0]
            if frag1 != prev_frag:
                prev_frag = frag1
                k = 0
            frag2 = data[i, 1]
            distance = mids[frag2] - mids[frag1]
            if distance < mindistance or distance > maxdistance:
                continue
            index = 0
            for j in range(num_features):
                bin1 = min(all_indices[frag1, j], all_indices[frag2, j])
                bin2 = max(all_indices[frag1, j], all_indices[frag2, j])
                index += ((num_bins[j] - 1) * bin1 - bin1 * (bin1 - 1) / 2 + bin2) * bin_divs[j]
            counts[index] += 1
            log_dist = log(<double>distance)
            signal = region_means[regions[frag1]] - log_dist * gamma
            log_count = log(<double>data[i, 2])
            if not corrections is None:
                log_count -= corrections[frag1] - corrections[frag2]
            sums[index, 0] += log_count - signal
            sums[index, 1] += pow(log_count - signal, 2.0)
        for i in range(num_trans):
            frag1 = trans_data[i, 0]
            frag2 = trans_data[i, 1]
            index = 0
            for j in range(num_features):
                bin1 = min(all_indices[frag1, j], all_indices[frag2, j])
                bin2 = max(all_indices[frag1, j], all_indices[frag2, j])
                index += ((num_bins[j] - 1) * bin1 - bin1 * (bin1 - 1) / 2 + bin2) * bin_divs[j]
            counts[index] += 1
            log_count = log(<double>trans_data[i, 2])
            if not corrections is None:
                log_count -= corrections[frag1] - corrections[frag2]
            sums[index, 0] += log_count - trans_mean
            sums[index, 1] += pow(log_count - trans_mean, 2.0)
    return None



#!/usr/bin/env python

import sys

try:
    from mpi4py import MPI
except:
    pass

from ..hic import HiC


def run(args):
    if 'mpi4py' in sys.modules.keys():
        comm = MPI.COMM_WORLD
        rank = comm.Get_rank()
        num_procs = comm.Get_size()
    else:
        comm = None
        rank = 0
        num_procs = 1
    if args.chroms is None:
        chroms = []
    else:
        chroms = args.chroms.split(',')
        if len(chroms) == 1 and chroms[0] == '':
            chroms = []
    if args.algorithm.count('binning') > 0:
        model = args.model.split(',')
        modelbins = args.modelbins.split(',')
        parameters = args.parameters.split(',')
        for i in range(len(modelbins)):
            try:
                modelbins[i] = int(modelbins[i])
            except:
                if rank == 0:
                    print sys.stderr, ("Not all arguments in -n/--modelbins could be converted to integers.")
                return 1
        if len(model) != len(modelbins):
            if rank == 0:
                print sys.stderr, ("-v/--model, -n/--modelbins, and -u/--parameter-types must be equal lengths.")
            return 1
    hic = HiC(args.project, 'r', silent=args.silent)
    precorrect = False
    if args.algorithm in ['binning', 'binning-express', 'binning-probability']:
        hic.find_binning_fend_corrections(mindistance=args.mindist, maxdistance=args.maxdist,
                                          chroms=chroms, num_bins=modelbins, model=model, parameters=parameters,
                                          usereads=args.binreads, learning_threshold=args.threshold,
                                          max_iterations=args.biniter, pseudocounts=args.pseudo)
        precorrect = True
    if args.algorithm in ['probability', 'binning-probability']:
        hic.find_probability_fend_corrections(mindistance=args.mindist, maxdistance=args.maxdist,
                                              minchange=args.change, max_iterations=args.probiter,
                                              learningstep=args.step, chroms=chroms,
                                              precalculate=args.precalc, precorrect=precorrect,
                                              model=args.probmodel)
    elif args.algorithm in ['express', 'binning-express']:
        hic.find_express_fend_corrections(iterations=args.expiter, mindistance=args.mindist,
                                          maxdistance=args.maxdist, remove_distance=args.nodist,
                                          usereads=args.expreads, mininteractions=args.minint,
                                          chroms=chroms, minchange=args.change, precorrect=precorrect,
                                          binary=args.binary, kr=args.kr)
    if rank == 0:
        hic.save(args.output)



#!/usr/bin/env python

import sys
import subprocess

import h5py

from ..hic import HiC


def run(args):
    if args.chroms is None:
        chroms = None
    else:
        chroms = args.chroms.split(',')
        if len(chroms) == 1 and chroms[0] == '':
            chroms = None
    hic = HiC(args.project, 'r', silent=args.silent)
    hic.write_multiresolution_heatmap(args.output, datatype=args.datatype, maxbinsize=args.maxbin,
                                      minbinsize=args.minbin, trans_maxbinsize=args.maxtransbin,
                                      trans_minbinsize=args.mintransbin, minobservations=args.minobs,
                                      chroms=chroms, includetrans=args.trans, midbinsize=args.midbin)



#!/usr/bin/env python

import os
import sys
import subprocess
import unittest
from operator import mul

import numpy

from hifive import hic
import h5py


class HiCProject(unittest.TestCase):
    def setUp(self):
        self.data_fname = 'test/data/test.hcd'
        self.bin_data_fname = 'test/data/test_bin.hcd'
        self.project_fname = 'test/data/test.hcp'
        self.bin_project_fname = 'test/data/test_bin.hcp'
        self.probbin_fname = 'test/data/test_probbin.hcp'
        self.probpois_fname = 'test/data/test_probpois.hcp'
        self.express_fname = 'test/data/test_express.hcp'
        self.bin_express_fname = 'test/data/test_bin_express.hcp'
        self.binning_fname = 'test/data/test_binning.hcp'
        self.data = h5py.File(self.project_fname, 'r')
        self.bin_data = h5py.File(self.bin_project_fname, 'r')
        self.probbin = hic.HiC(self.probbin_fname, 'r')
        self.probpois = hic.HiC(self.probpois_fname, 'r')
        self.express = hic.HiC(self.express_fname, 'r')
        self.bin_express = hic.HiC(self.bin_express_fname, 'r')
        self.binning = hic.HiC(self.binning_fname, 'r')

    def test_hic_project_preanalysis(self):
        subprocess.call("./bin/hifive hic-project -q -m 20000 -f 10 -j 30000 -n 5 %s test/data/test_temp.hcp" %
                        (self.data_fname), shell=True)
        project = h5py.File('test/data/test_temp.hcp', 'r')
        self.compare_hdf5_dicts(self.data, project, 'project')

    def test_hic_bin_project_preanalysis(self):
        subprocess.call("./bin/hifive hic-project -q -m 20000 -f 10 -j 30000 -n 5 %s test/data/test_temp.hcp" %
                        (self.bin_data_fname), shell=True)
        project = h5py.File('test/data/test_temp.hcp', 'r')
        self.compare_hdf5_dicts(self.bin_data, project, 'project')

    def test_hic_project_probability_binomial(self):
        subprocess.call("./bin/hifive hic-normalize probability -q -m 20000 -o test/data/test_temp.hcp -b 15 -l 0.4 -g 0.0015 -p %s" %
                        (self.project_fname), shell=True)
        project = hic.HiC("test/data/test_temp.hcp", 'r', silent=True)
        self.assertTrue(numpy.allclose(self.probbin.corrections, project.corrections, atol=1e-4),
            "learned correction values don't match target values")
        self.assertTrue(numpy.allclose(self.probbin.chromosome_means, project.chromosome_means, atol=1e-4),
            "chromosome means don't match target values")

    def test_hic_project_probability_poisson(self):
        subprocess.call("./bin/hifive hic-normalize probability -q -m 20000 -o test/data/test_temp.hcp -b 15 -l 0.4 -g 0.0015 -p -a poisson %s" %
                        (self.project_fname), shell=True)
        project = hic.HiC("test/data/test_temp.hcp", 'r', silent=True)
        self.assertTrue(numpy.allclose(self.probpois.corrections, project.corrections, atol=1e-1),
            "learned correction values don't match target values")
        self.assertTrue(numpy.allclose(self.probpois.chromosome_means, project.chromosome_means, atol=1e-4),
            "chromosome means don't match target values")

    def test_hic_project_express(self):
        subprocess.call("./bin/hifive hic-normalize express -q -m 20000 -o test/data/test_temp.hcp -e 100 -w cis -f 10 %s" %
                        (self.project_fname), shell=True)
        project = hic.HiC("test/data/test_temp.hcp", 'r', silent=True)
        self.assertTrue(numpy.allclose(self.express.corrections, project.corrections),
            "learned express correction values don't match target values")
        self.assertTrue(numpy.allclose(self.express.chromosome_means, project.chromosome_means),
            "chromosome means don't match target values")

    def test_hic_project_binning(self):
        subprocess.call("./bin/hifive hic-normalize binning -q -m 20000 -o test/data/test_temp.hcp -r 5 -y cis -t 1.0 -v len,distance -s 3,3 -u even,fixed-const %s" %
                        (self.project_fname), shell=True)
        project = hic.HiC("test/data/test_temp.hcp", 'r', silent=True)
        self.assertTrue(numpy.allclose(self.binning.binning_corrections, project.binning_corrections),
            "learned binning correction values don't match target values")

    def test_hic_bin_project_express(self):
        subprocess.call("./bin/hifive hic-normalize express -q -m 20000 -o test/data/test_temp.hcp -e 100 -w cis -f 10 %s" %
                        (self.bin_project_fname), shell=True)
        project = hic.HiC("test/data/test_temp.hcp", 'r', silent=True)
        self.assertTrue(numpy.allclose(self.bin_express.corrections, project.corrections),
            "learned express correction values don't match target values")
        self.assertTrue(numpy.allclose(self.bin_express.chromosome_means, project.chromosome_means),
            "chromosome means don't match target values")

    def tearDown(self):
        subprocess.call('rm -f test/data/test_temp.hcp', shell=True)

    def compare_arrays(self, array1, array2, name):
        self.assertTrue(array1.shape == array2.shape,
            "%s shape doesn't match target value" % name)
        self.assertTrue(array1.dtype == array2.dtype,
            "%s dtype doesn't match target value" % name)
        if str(array1.dtype).count('S') + str(array1.dtype).count('a') > 0:
            self.assertTrue(numpy.sum(array1 == array2) == reduce(mul, array1.shape),
                "%s don't match target values." % name)
        else:
            self.assertTrue(numpy.allclose(array1, array2),
                "%s don't match target values" % name)
        return None

    def compare_hdf5_dicts(self, dict1, dict2, name):
        for key in dict1['/'].attrs.keys():
            if key == 'history':
                continue
            self.assertTrue(key in dict2['/'].attrs,
                "%s missing from %s attributes" % (key, name))
            value1 = dict1['/'].attrs[key]
            value2 = dict2['/'].attrs[key]
            if isinstance(value1, float):
                self.assertTrue(
                    numpy.allclose(numpy.array([value1], dtype=numpy.float32), numpy.array([value2], dtype=numpy.float32)),
                    "%s in %s doesn't match target value" % (key, name))
            elif isinstance(value1, int):
                self.assertTrue(
                    numpy.allclose(numpy.array([value1], dtype=numpy.int32), numpy.array([value2], dtype=numpy.int32)),
                    "%s in %s doesn't match target value" % (key, name))
            else:
                self.assertTrue(value1 == value2, "%s in %s doesn't match target value" % (key, name))
        for key in dict1.keys():
            self.assertTrue(key in dict2,
                "%s missing from heatmap arrays" % name)
            self.compare_arrays(dict1[key][...], dict2[key][...], name)
        return None


if __name__ == "__main__":
    unittest.main()


#!/usr/bin/env python

import os
import sys
import subprocess
import unittest
from operator import mul

import numpy

from hifive import fivec
import h5py


class FiveCProject(unittest.TestCase):
    def setUp(self):
        self.data_fname = 'test/data/test.fcd'
        self.project_fname = 'test/data/test.fcp'
        self.probability_fname = 'test/data/test_probability.fcp'
        self.express_fname = 'test/data/test_express.fcp'
        self.binning_fname = 'test/data/test_binning.fcp'
        self.raw = h5py.File(self.project_fname, 'r')
        self.probability = fivec.FiveC(self.probability_fname, 'r')
        self.express = fivec.FiveC(self.express_fname, 'r')
        self.binning = fivec.FiveC(self.binning_fname, 'r')

    def test_fivec_project_preanalysis(self):
        subprocess.call("./bin/hifive 5c-project -q -f 20 %s test/data/test_temp.fcp" %
                        self.data_fname, shell=True)
        project = h5py.File('test/data/test_temp.fcp', 'r')
        self.compare_hdf5_dicts(self.raw, project, 'project')

    def test_fivec_project_probability(self):
        subprocess.call("./bin/hifive 5c-normalize probability -q -o test/data/test_temp.fcp -b 100 -l 0.5 -p %s" %
                        self.project_fname, shell=True)
        project = fivec.FiveC("test/data/test_temp.fcp", 'r', silent=True)
        self.assertTrue(numpy.allclose(self.probability.corrections, project.corrections),
            "learned correction values don't match target values")
        self.assertTrue(numpy.allclose(self.probability.region_means, project.region_means),
                                       "region means don't match target values")

    def test_fivec_project_express(self):
        subprocess.call("./bin/hifive 5c-normalize express -q -o test/data/test_temp.fcp -e 100 -d -w cis %s" %
                        self.project_fname, shell=True)
        project = fivec.FiveC("test/data/test_temp.fcp", 'r', silent=True)
        self.assertTrue(numpy.allclose(self.express.corrections, project.corrections),
            "learned express correction values don't match target values")
        self.assertTrue(numpy.allclose(self.express.region_means, project.region_means),
                                       "region means don't match target values")

    def test_fivec_project_binning(self):
        subprocess.call("./bin/hifive 5c-normalize binning -q -o test/data/test_temp.fcp -i 10 -t 1.0 -y cis -v len -n 5 -u even %s" %
                        self.project_fname, shell=True)
        project = fivec.FiveC("test/data/test_temp.fcp", 'r', silent=True)
        self.assertTrue(numpy.allclose(self.binning.binning_corrections, project.binning_corrections),
            "learned binning correction values don't match target values")

    def tearDown(self):
        subprocess.call('rm -f test/data/test_temp.fcp', shell=True)

    def compare_arrays(self, array1, array2, name):
        self.assertTrue(array1.shape == array2.shape,
            "%s shape doesn't match target value" % name)
        self.assertTrue(array1.dtype == array2.dtype,
            "%s dtype doesn't match target value" % name)
        if str(array1.dtype).count('S') + str(array1.dtype).count('a') > 0:
            self.assertTrue(numpy.sum(array1 == array2) == reduce(mul, array1.shape),
                "%s don't match target values." % name)
        else:
            self.assertTrue(numpy.allclose(array1, array2),
                "%s don't match target values" % name)
        return None

    def compare_hdf5_dicts(self, dict1, dict2, name):
        for key in dict1['/'].attrs.keys():
            if key == 'history':
                continue
            self.assertTrue(key in dict2['/'].attrs,
                "%s missing from %s attributes" % (key, name))
            value1 = dict1['/'].attrs[key]
            value2 = dict2['/'].attrs[key]
            if isinstance(value1, float):
                self.assertTrue(
                    numpy.allclose(numpy.array([value1], dtype=numpy.float32), numpy.array([value2], dtype=numpy.float32)),
                    "%s in %s doesn't match target value" % (key, name))
            elif isinstance(value1, int):
                self.assertTrue(
                    numpy.allclose(numpy.array([value1], dtype=numpy.int32), numpy.array([value2], dtype=numpy.int32)),
                    "%s in %s doesn't match target value" % (key, name))
            else:
                self.assertTrue(value1 == value2, "%s in %s doesn't match target value" % (key, name))
        for key in dict1.keys():
            self.assertTrue(key in dict2,
                "%s missing from heatmap arrays" % name)
            self.compare_arrays(dict1[key][...], dict2[key][...], name)
        return None


if __name__ == "__main__":
    unittest.main()


#!/usr/bin/env python

import sys
import subprocess

import h5py
try:
    from pyx import *
except:
    pass

from ..fivec import FiveC
from ..plotting import plot_fivec_dict, plot_key


def run(args):
    if not args.image is None and args.pdf and "pyx" not in sys.modules.keys():
        print >> sys.stderr, ("-p/--pdf requires the package 'pyx'"),
        return 1
    if args.binsize > 0:
        args.arraytype = 'full'
    if args.regions is None:
        regions = []
    else:
        regions = args.regions.split(',')
        if len(regions) == 1 and regions[0] == '':
            regions = []
    for i in range(len(regions)):
        try:
            regions[i] = int(regions[i])
        except:
            print sys.stderr, ("Not all arguments in -r/--regions could be converted to integers.")
            return 1
    fivec = FiveC(args.project, 'r', silent=args.silent)
    heatmaps = fivec.write_heatmap(args.output, binsize=args.binsize, includetrans=args.trans,
                        arraytype=args.arraytype, datatype=args.datatype, regions=regions,
                        dynamically_binned=args.dynamic, expansion_binsize=args.expbinsize,
                        minobservations=args.minobs, searchdistance=args.search,
                        removefailed=args.remove, format=args.format)
    if not args.image is None:
        kwargs = {}
        for arg in args.keywords:
            temp = arg.split("=")
            if temp[1] in ["True", "TRUE", "true"]:
                temp[1] = True
            elif temp[1] in ["False", "FALSE", "false"]:
                temp[1] = False
            elif temp[1][0] == "(":
                temp[1] = temp[1].strip('()').split(',')
                for i in range(len(temp[1])):
                    temp[1][i] = int(temp[1][i])
                temp[1] = tuple(temp[1])
            elif temp[1][0] == "[":
                temp[1] = temp[1].strip('[]').split(',')
                for i in range(len(temp[1])):
                    temp[1][i] = int(temp[1][i])
            else:
                try:
                    temp[1] = int(temp[1])
                except:
                    try:
                        temp[1] = float(temp[1])
                    except:
                        # strip off extra characters introduced by galaxy into color format
                        temp[1] = temp[1].replace('__pd__','')
            kwargs[temp[0]] = temp[1]
        if 'symmetricscaling' not in kwargs:
            if args.datatype == 'enrichment':
                kwargs['symmetricscaling'] = True
            else:
                kwargs['symmetricscaling'] = False
        img, minscore, maxscore = plot_fivec_dict(heatmaps, returnscale=True, silent=args.silent, **kwargs)
        if not args.pdf:
            img_format = args.image.split('.')[-1].upper()
            if img_format not in ['PNG', 'TIF', 'JPG', 'JPEG']:
                img_format = 'PNG'
            img.save(args.image, img_format)
        else:
            unit.set(defaultunit="cm")
            text.set(mode="latex")
            text.preamble(r"\usepackage{times}")
            text.preamble(r"\usepackage{sansmath}")
            text.preamble(r"\sansmath")
            text.preamble(r"\renewcommand*\familydefault{\sfdefault}")
            hm = h5py.File(args.output, 'r')
            totalx = 0
            totaly = 0
            minsize = 999999999
            sizes = [0]
            names = []
            all_regions = hm['regions'][...]
            for i, region in enumerate(all_regions['index']):
                names.append("chr%s:%i-%i" % (all_regions['chromosome'][i], all_regions['start'][i],
                                              all_regions['stop'][i]))
                if "%i.fragments" % region in hm:
                    name = "%i.fragments" % int(region)
                    sizes.append(hm[name].shape[0])
                    totalx += hm[name].shape[0]
                    totaly += hm[name].shape[0]
                    minsize = min(minsize, hm[name].shape[0])
                elif "%i.reverse_fragments" % int(region) in hm:
                    name = "%i.reverse_fragments" % int(region)
                    sizes.append(hm[name].shape[0])
                    totaly += hm[name].shape[0]
                    totalx += hm["%i.forward_fragments" % int(region)].shape[0]
                    minsize = min(minsize, hm[name].shape[0])
                elif "%i.positions" % int(region) in hm:
                    name = "%i.positions" % int(region)
                    sizes.append(hm[name].shape[0])
                    totalx += hm[name].shape[0]
                    totaly += hm[name].shape[0]
                    minsize = min(minsize, hm[name].shape[0])
            totaly += all_regions.shape[0] - 1
            totalx += all_regions.shape[0] - 1
            height = max(5.0, totaly * 0.5 / minsize)
            width = height / totaly * totalx
            if len(sizes) > 2:
                sizes[1] += 0.5
                sizes[-1] += 0.5
                if len(sizes) > 3:
                    for i in range(2, len(sizes) - 1):
                        sizes[i] += 1.0
            for i in range(1, len(sizes)):
                sizes[i] += sizes[i - 1]
            for i in range(1, len(sizes)):
                sizes[i] = sizes[i] / totaly * height
            c = canvas.canvas()
            c.insert(bitmap.bitmap(0, 0, img, width=width))
            if args.legend:
                if 'min_color' in kwargs:
                    min_color = kwargs['min_color']
                else:
                    min_color = (0, 0, 1)
                if 'mid_color' in kwargs:
                    mid_color = kwargs['mid_color']
                else:
                    mid_color = (1, 1, 1)
                if 'max_color' in kwargs:
                    max_color = kwargs['max_color']
                else:
                    max_color = (1, 0, 0)
                if 'logged' in kwargs:
                    logged = kwargs['logged']
                else:
                    logged = True
                c.insert(plot_key(min_score=minscore, max_score=maxscore, height=(height * 0.05),
                                  width=width, orientation='top', num_ticks=5, min_color=min_color,
                                  mid_color=mid_color, max_color=max_color, log_display=False),
                                  [trafo.translate(0, height * 1.05)])
                if logged:
                    label = "Log2 "
                else:
                    label = ""
                if args.datatype == 'enrichment':
                    c.text(width * 0.5, height * 1.1 + 0.75, "%sEnrichment" % (label),
                           [text.halign.center, text.valign.bottom, text.size(-2)])
                elif args.datatype == 'raw':
                    c.text(width * 0.5, height * 1.1 + 0.75, "%sCounts" % (label),
                           [text.halign.center, text.valign.bottom, text.size(-2)])
                else:
                    c.text(width * 0.5, height * 1.1 + 0.75, "%sNormalized Counts" % (label),
                           [text.halign.center, text.valign.bottom, text.size(-2)])
            if args.names:
                for i, name in enumerate(names):
                    c.text(width + 0.25, height - (sizes[i] + sizes[i + 1]) / 2, name,
                           [text.halign.left, text.valign.middle, text.size(-2)])
            c.writePDFfile(args.image)
            if len(args.image.split('.')) <= 1 or args.image.split('.')[-1] != 'pdf':
                subprocess.call('mv %s.pdf %s' % (args.image, args.image), shell=True)



#!/usr/bin/env python

import sys
import subprocess
from math import floor, ceil, log10

import numpy
try:
    from pyx import *
except:
    pass

from ..fivec import FiveC
from ..plotting import plot_full_array, plot_upper_array, plot_diagonal_from_upper_array, plot_key


def run(args):
    if not args.image is None and args.pdf and "pyx" not in sys.modules.keys():
        print sys.stderr, ("-p/--pdf requires the package 'pyx'"),
        return 1
    fivec = FiveC(args.project, 'r', silent=args.silent)
    if args.binsize == 0:
        arraytype = 'compact'
    else:
        arraytype = 'upper'
    kwargs = {}
    for arg in args.keywords:
        temp = arg.split("=")
        if temp[1] in ["True", "TRUE", "true"]:
            temp[1] = True
        elif temp[1] in ["False", "FALSE", "false"]:
            temp[1] = False
        elif temp[1][0] == "(":
            temp[1] = temp[1].strip('()').split(',')
            for i in range(len(temp[1])):
                temp[1][i] = int(temp[1][i])
            temp[1] = tuple(temp[1])
        elif temp[1][0] == "[":
            temp[1] = temp[1].strip('[]').split(',')
            for i in range(len(temp[1])):
                temp[1][i] = int(temp[1][i])
        else:
            try:
                temp[1] = int(temp[1])
            except:
                # strip off extra characters introduced by galaxy into color format
                temp[1] = temp[1].replace('__pd__','')
    chrom = fivec.frags['regions']['chromosome'][args.region]
    if args.region2 is None:
        temp = fivec.cis_heatmap(region=args.region, binsize=args.binsize, start=args.start,
                                 stop=args.stop, datatype=args.datatype, arraytype=arraytype,
                                 returnmapping=True, skipfiltered=True, dynamically_binned=args.dynamic,
                                 expansion_binsize=args.expbinsize, minobservations=args.minobs,
                                 searchdistance=args.search, removefailed=args.remove)
    else:
        chrom2 = fivec.frags['regions']['chromosome'][args.region2]
        temp = fivec.trans_heatmap(region1=args.region, region2=args.region2, binsize=args.binsize, start1=args.start,
                                 stop1=args.stop, start2=args.start2, stop2=args.stop2, datatype=args.datatype,
                                 arraytype='full', returnmapping=True, skipfiltered=True,
                                 dynamically_binned=args.dynamic, expansion_binsize=args.expbinsize,
                                 minobservations=args.minobs, searchdistance=args.search, removefailed=args.remove)
    output = open(args.output, 'w')
    if args.region2 is None:
        if args.binsize == 0:
            data = temp[0]
            xmapping = temp[1][:, :2]
            ymapping = temp[2][:, :2]
            all_data = []
            for i in range(xmapping.shape[0]):
                for j in range(ymapping.shape[0]):
                    if data[i, j, 0] <= 0.0:
                        continue
                    if xmapping[i, 0] < ymapping[j, 0]:
                        all_data.append((xmapping[i, 0], xmapping[i, 1], ymapping[j, 0], ymapping[j, 1],
                                         numpy.log2(data[i, j, 0] / data[i, j, 1])))
                    else:
                        all_data.append((ymapping[j, 0], ymapping[j, 1], xmapping[i, 0], xmapping[i, 1],
                                         numpy.log2(data[i, j, 0] / data[i, j, 1])))
            all_data = numpy.array(all_data, dtype=numpy.dtype([('start1', numpy.int32), ('stop1', numpy.int32),
                                                                ('start2', numpy.int32), ('stop2', numpy.int32),
                                                                ('value', numpy.float32)]))
            order = numpy.lexsort((all_data['start2'], all_data['start1']))
            for i in order:
                print >> output, "chr%s\t%i\t%i\tchr%s\t%i\t%i\t%f" % (chrom, all_data['start1'][i],
                    all_data['stop1'][i], chrom, all_data['start2'][i], all_data['stop2'][i], all_data['value'][i])
        else:
            data = temp[0]
            mapping = temp[1][:, :2]
            pos = 0
            for i in range(mapping.shape[0] - 1):
                for j in range(i + 1, mapping.shape[0]):
                    if data[pos, 0] > 0.0 and data[pos, 1] > 0.0:
                        print >> output, "chr%s\t%i\t%i\tchr%s\t%i\t%i\t%f" % (chrom, mapping[i, 0], mapping[i, 1],
                            chrom, mapping[j, 0], mapping[j, 1], numpy.log2(data[pos, 0] / data[pos, 1]))
                    pos += 1
    else:
        data, mapping1, mapping2 = temp
        for i in range(mapping1.shape[0]):
            for j in range(mapping2.shape[0]):
                if data[i, j, 0] <= 0.0:
                    continue
                print >> output, "chr%s\t%i\t%i\tchr%s\t%i\t%i\t%f" % (chrom, mapping1[i, 0], mapping1[i, 1], chrom2,
                                                                       mapping2[j, 0], mapping2[j, 1],
                                                                       numpy.log2(data[i, j, 0] / data[i, j, 1]))
    output.close()
    if not args.image is None:
        if args.datatype == 'enrichment':
            symmetricscaling = True
        else:
            symmetricscaling = False
        if 'symmetricscaling' in kwargs:
            symmetricscaling = kwargs['symmetricscaling']
        if not args.region2 is None:
            img, minscore, maxscore = plot_full_array(data, returnscale=True, symmetricscaling=symmetricscaling,
                                                      silent=args.silent, **kwargs)
            offset = 0.0
            width = 5.0
            height = width / mapping1.shape[0] * mapping2.shape[0]
        elif arraytype == 'compact':
            img, minscore, maxscore = plot_full_array(data, returnscale=True, symmetricscaling=symmetricscaling,
                                                      silent=args.silent, **kwargs)
            offset = 0.0
            width = 5.0
            height = width / xmapping.shape[0] * ymapping.shape[0]
        else:
            if args.rotate:
                img, minscore, maxscore = plot_diagonal_from_upper_array(data, returnscale=True,
                                          symmetricscaling=symmetricscaling, silent=args.silent, **kwargs)
                offset = 2.5 / (mapping.shape[0] * 2 - 2)
                height = 2.5
                width = 5.0
            else:
                img, minscore, maxscore = plot_upper_array(data, returnscale=True, symmetricscaling=symmetricscaling,
                                                           silent=args.silent, **kwargs)
                offset = 0.0
                height = width = 5.0
        if args.pdf:
            c = canvas.canvas()
            c1 = canvas.canvas([canvas.clip(path.rect(0, 0, width, height))])
            c1.insert(bitmap.bitmap(-offset, -offset, img, width=width))
            c.insert(c1)
            if args.region2 is None:
                if args.ticks and args.binsize > 0:
                    c.stroke(path.line(0, 0, width, 0))
                    xmin = (mapping[0, 0] + mapping[0, 1]) / 2
                    xmax = (mapping[-1, 0] + mapping[-1, 1]) / 2
                    order = int(floor(log10(xmax - xmin))) - 1
                    step = int(floor((xmax - xmin) / (10.0 ** order * width))) * 10 ** order
                    values = numpy.arange(((xmin - 1) / step + 1) * step, (xmax / step) * step + 1, step)
                    ticks = (values - float(mapping[0, 0] + mapping[0, 1]) / 2) / (mapping[-1, 0] -
                                                                                   mapping[0, 0]) * width
                    for i in range(values.shape[0]):
                        c.stroke(path.line(ticks[i], 0, ticks[i], -0.25), [style.linewidth.Thin])
                        c.text(ticks[i], -0.3, "%0.2e" % values[i],
                               [text.valign.middle, text.halign.left, text.size(-2), trafo.rotate(-90)])
                    if not args.rotate:
                        c.stroke(path.line(width, 0, width, height))
                        for i in range(values.shape[0]):
                            c.stroke(path.line(width, height - ticks[i], width + 0.25, height - ticks[i]),
                                     [style.linewidth.Thin])
                            c.text(width + 0.3, height - ticks[i], "%0.2e" % values[i],
                                   [text.valign.middle, text.halign.left, text.size(-2)])
            elif args.ticks:
                c.stroke(path.line(0, 0, width, 0))
                xmin = (mapping1[0, 0] + mapping1[0, 1]) / 2
                xmax = (mapping1[-1, 0] + mapping1[-1, 1]) / 2
                order = int(floor(log10(xmax - xmin))) - 1
                step = int(floor((xmax - xmin) / (10.0 ** order * width))) * 10 ** order
                values = numpy.arange(((xmin - 1) / step + 1) * step, (xmax / step) * step + 1, step)
                ticks = (values - float(mapping1[0, 0] + mapping1[0, 1]) / 2) / (mapping1[-1, 0] -
                                                                               mapping1[0, 0]) * width
                for i in range(values.shape[0]):
                    c.stroke(path.line(ticks[i], 0, ticks[i], -0.25), [style.linewidth.Thin])
                    c.text(ticks[i], -0.3, "%0.2e" % values[i],
                           [text.valign.middle, text.halign.left, text.size(-2), trafo.rotate(-90)])

                c.stroke(path.line(width, 0, width, height))
                xmin = (mapping2[0, 0] + mapping2[0, 1]) / 2
                xmax = (mapping2[-1, 0] + mapping2[-1, 1]) / 2
                order = int(floor(log10(xmax - xmin))) - 1
                step = int(floor((xmax - xmin) / (10.0 ** order * width))) * 10 ** order
                values = numpy.arange(((xmin - 1) / step + 1) * step, (xmax / step) * step + 1, step)
                ticks = (values - float(mapping2[0, 0] + mapping2[0, 1]) / 2) / (mapping2[-1, 0] -
                                                                               mapping2[0, 0]) * height
                for i in range(values.shape[0]):
                    c.stroke(path.line(width, height - ticks[i], width + 0.25, height - ticks[i]),
                             [style.linewidth.Thin])
                    c.text(width + 0.3, height - ticks[i], "%0.2e" % values[i],
                           [text.valign.middle, text.halign.left, text.size(-2)])
            if args.legend:
                if 'min_color' in kwargs:
                    min_color = kwargs['min_color']
                else:
                    min_color = "0000ff"
                if 'mid_color' in kwargs:
                    mid_color = kwargs['mid_color']
                else:
                    mid_color = "ffffff"
                if 'max_color' in kwargs:
                    max_color = kwargs['max_color']
                else:
                    max_color = "ff0000"
                if 'logged' in kwargs:
                    logged = kwargs['logged']
                else:
                    logged = True
                c.insert(plot_key(min_score=minscore, max_score=maxscore, height=0.25, width=width,
                                  orientation='top', num_ticks=5, min_color=min_color,
                                  mid_color=mid_color, max_color=max_color,
                                  log_display=False), [trafo.translate(0, height + 0.25)])
                if logged:
                    label = "Log2 "
                else:
                    label = ""
                if args.datatype == 'enrichment':
                    c.text(width * 0.5, height + 0.8, "%sEnrichment" % label, [text.halign.center, text.valign.bottom,
                                                                       text.size(-2)])
                elif args.datatype == 'raw':
                    c.text(width + 0.5, height + 0.8, "%sCounts" % label, [text.halign.center, text.valign.bottom,
                                                                   text.size(-2)])
                else:
                    c.text(width * 0.5, height + 0.8, "%sNormalized Counts" % label,
                           [text.halign.center, text.valign.bottom, text.size(-2)])

            c.writePDFfile(args.image)
            if len(args.image.split('.')) <= 1 or args.image.split('.')[-1] != 'pdf':
                subprocess.call('mv %s.pdf %s' % (args.image, args.image), shell=True)

        else:
            img_format = args.image.split('.')[-1].upper()
            if img_format not in ['PNG', 'TIF', 'JPG', 'JPEG']:
                img_format = 'PNG'
            img.save(args.image, img_format)



#!/usr/bin/env python

import sys

try:
    from mpi4py import MPI
except:
    pass

from ..hic import HiC


def run(args):
    if 'mpi4py' in sys.modules.keys():
        comm = MPI.COMM_WORLD
        rank = comm.Get_rank()
        num_procs = comm.Get_size()
    else:
        comm = None
        rank = 0
        num_procs = 1
    if rank == 0:
        hic = HiC(args.output, 'w', silent=args.silent)
        hic.load_data(args.data)
        hic.filter_fends(mininteractions=args.minint, mindistance=args.mindist, maxdistance=args.maxdist)
        hic.save()
        for i in range(1, num_procs):
            comm.send(1, dest=i, tag=11)
    else:
        comm.recv(source=0, tag=11)
        hic = HiC(args.output, 'r', silent=True)
    hic.find_distance_parameters(minsize=args.minbin, numbins=args.numbins)
    if rank == 0:
        hic.save()
    return None



# distutils: language = c++

"""These functions provide increased speed in handling functions dealing with correction
parameter optimization necessary for supporting hic analysis using HiFive.
"""

import cython
cimport numpy as np
import numpy

ctypedef np.float32_t DTYPE_t
ctypedef np.float64_t DTYPE_64_t
ctypedef np.int32_t DTYPE_int_t
ctypedef np.int64_t DTYPE_int64_t
ctypedef np.uint32_t DTYPE_uint_t
ctypedef np.int8_t DTYPE_int8_t
cdef double Inf = numpy.inf

cdef extern from "math.h":
    double exp(double x) nogil
    double log(double x) nogil
    double log10(double x) nogil
    double sqrt(double x) nogil
    double pow(double x, double x) nogil
    double abs(double x) nogil
    double round(double x) nogil
    double floor(double x) nogil
    double ceil(double x) nogil


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def find_binning_correction_adjustment(
        np.ndarray[DTYPE_t, ndim=1] corrections not None,
        np.ndarray[DTYPE_int_t, ndim=1] indices0 not None,
        np.ndarray[DTYPE_int_t, ndim=1] indices1 not None,
        np.ndarray[DTYPE_t, ndim=1] binning_corrections,
        np.ndarray[DTYPE_int_t, ndim=1] num_bins,
        np.ndarray[DTYPE_int_t, ndim=3] fend_indices):
    cdef long long int i, j, fend1, fend2, index
    cdef long long int num_fends = corrections.shape[0]
    cdef long long int num_parameters = fend_indices.shape[1]
    with nogil:
        for i in range(num_fends):
            fend1 = indices0[i]
            fend2 = indices1[i]
            for j in range(num_parameters):
                if fend_indices[fend1, j, 0] < fend_indices[fend2, j, 0]:
                    corrections[i] *= binning_corrections[fend_indices[fend1, j, 1] + fend_indices[fend2, j, 0]]
                else:
                    corrections[i] *= binning_corrections[fend_indices[fend2, j, 1] + fend_indices[fend1, j, 0]]
    return None


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def calculate_binom_gradients(
        np.ndarray[DTYPE_int_t, ndim=1] counts,
        np.ndarray[DTYPE_int_t, ndim=1] zero_indices0 not None,
        np.ndarray[DTYPE_int_t, ndim=1] zero_indices1 not None,
        np.ndarray[DTYPE_int_t, ndim=1] nonzero_indices0 not None,
        np.ndarray[DTYPE_int_t, ndim=1] nonzero_indices1 not None,
        np.ndarray[DTYPE_t, ndim=1] nonzero_means,
        np.ndarray[DTYPE_t, ndim=1] zero_means not None,
        np.ndarray[DTYPE_t, ndim=1] corrections not None,
        np.ndarray[DTYPE_t, ndim=1] inv_corrections not None,
        np.ndarray[DTYPE_64_t, ndim=1] gradients not None):
    cdef long long int i, index0, index1
    cdef double value, distance_mean, correction0, correction1
    cdef long long int num_zero_pairs = zero_indices0.shape[0]
    cdef long long int num_nonzero_pairs = nonzero_indices0.shape[0]
    with nogil:
        for i in range(num_nonzero_pairs):
            index0 = nonzero_indices0[i]
            index1 = nonzero_indices1[i]
            gradients[index0] -= inv_corrections[index0]
            if index1 != index0:
                gradients[index1] -= inv_corrections[index1]
        for i in range(num_zero_pairs):
            index0 = zero_indices0[i]
            index1 = zero_indices1[i]
            distance_mean = zero_means[i]
            correction0 = corrections[index0]
            correction1 = corrections[index1]
            value = 1.0 / (1.0 - distance_mean * corrections[index0] * corrections[index1])
            gradients[index0] += (distance_mean * corrections[index1]) * value
            if index1 != index0:
                gradients[index1] += (distance_mean * corrections[index0]) * value
    return None


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def calculate_binom_cost(
        np.ndarray[DTYPE_int_t, ndim=1] counts,
        np.ndarray[DTYPE_int_t, ndim=1] zero_indices0 not None,
        np.ndarray[DTYPE_int_t, ndim=1] zero_indices1 not None,
        np.ndarray[DTYPE_int_t, ndim=1] nonzero_indices0 not None,
        np.ndarray[DTYPE_int_t, ndim=1] nonzero_indices1 not None,
        np.ndarray[DTYPE_t, ndim=1] nonzero_means not None,
        np.ndarray[DTYPE_t, ndim=1] zero_means not None,
        np.ndarray[DTYPE_t, ndim=1] corrections not None,
        np.ndarray[DTYPE_t, ndim=1] log_corrections not None):
    cdef long long int i
    cdef double cost
    cdef long long int num_zero_pairs = zero_indices0.shape[0]
    cdef long long int num_nonzero_pairs = nonzero_indices0.shape[0]
    with nogil:
        cost = 0.0
        for i in range(num_nonzero_pairs):
            cost -= nonzero_means[i] + log_corrections[nonzero_indices0[i]] + log_corrections[nonzero_indices1[i]]
        for i in range(num_zero_pairs):
            cost -= log(max(0.0000001, 1.0 - zero_means[i] * corrections[zero_indices0[i]] * corrections[zero_indices1[i]]))
    return cost


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def calculate_poisson_gradients(
        np.ndarray[DTYPE_int_t, ndim=1] counts not None,
        np.ndarray[DTYPE_int_t, ndim=1] zero_indices0 not None,
        np.ndarray[DTYPE_int_t, ndim=1] zero_indices1 not None,
        np.ndarray[DTYPE_int_t, ndim=1] nonzero_indices0 not None,
        np.ndarray[DTYPE_int_t, ndim=1] nonzero_indices1 not None,
        np.ndarray[DTYPE_t, ndim=1] nonzero_means not None,
        np.ndarray[DTYPE_t, ndim=1] zero_means not None,
        np.ndarray[DTYPE_t, ndim=1] corrections not None,
        np.ndarray[DTYPE_t, ndim=1] inv_corrections not None,
        np.ndarray[DTYPE_64_t, ndim=1] gradients not None):
    cdef long long int i, index0, index1
    cdef long long int num_zero_pairs = zero_indices0.shape[0]
    cdef long long int num_nonzero_pairs = nonzero_indices0.shape[0]
    with nogil:
        for i in range(num_nonzero_pairs):
            index0 = nonzero_indices0[i]
            index1 = nonzero_indices1[i]
            gradients[index0] += nonzero_means[i] * corrections[index0] - counts[i] * inv_corrections[index0]
            if index1 != index0:
                gradients[index1] += nonzero_means[i] * corrections[index1] - counts[i] * inv_corrections[index1]
        for i in range(num_zero_pairs):
            index0 = zero_indices0[i]
            index1 = zero_indices1[i]
            gradients[index0] += zero_means[i] * corrections[index0]
            if index1 != index0:
                gradients[index1] += zero_means[i] * corrections[index1]
    return None


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def calculate_poisson_cost(
        np.ndarray[DTYPE_int_t, ndim=1] counts not None,
        np.ndarray[DTYPE_int_t, ndim=1] zero_indices0 not None,
        np.ndarray[DTYPE_int_t, ndim=1] zero_indices1 not None,
        np.ndarray[DTYPE_int_t, ndim=1] nonzero_indices0 not None,
        np.ndarray[DTYPE_int_t, ndim=1] nonzero_indices1 not None,
        np.ndarray[DTYPE_t, ndim=1] nonzero_means not None,
        np.ndarray[DTYPE_t, ndim=1] zero_means not None,
        np.ndarray[DTYPE_t, ndim=1] corrections not None,
        np.ndarray[DTYPE_t, ndim=1] log_corrections not None):
    cdef long long int i
    cdef double cost
    cdef long long int num_zero_pairs = zero_indices0.shape[0]
    cdef long long int num_nonzero_pairs = nonzero_indices0.shape[0]
    with nogil:
        cost = 0.0
        for i in range(num_nonzero_pairs):
            cost += corrections[nonzero_indices0[i]] * corrections[nonzero_indices1[i]] * nonzero_means[i] - counts[i] * ( log(nonzero_means[i]) + log_corrections[nonzero_indices0[i]] + log_corrections[nonzero_indices1[i]] )
        for i in range(num_zero_pairs):
            cost += corrections[zero_indices0[i]] * corrections[zero_indices1[i]] * zero_means[i]
    return cost


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def find_fend_means(
        np.ndarray[DTYPE_t, ndim=1] distance_means,
        np.ndarray[DTYPE_t, ndim=1] trans_means,
        np.ndarray[DTYPE_64_t, ndim=1] fend_means not None,
        np.ndarray[DTYPE_int_t, ndim=2] data,
        np.ndarray[DTYPE_int_t, ndim=2] trans_data,
        np.ndarray[DTYPE_t, ndim=1] corrections not None,
        double mu,
        double trans_mu,
        int binary):
    cdef long long int i, fend1, fend2, num_trans_data, num_data
    cdef double temp
    cdef long long int num_fends = fend_means.shape[0]
    if not trans_data is None:
        num_trans_data = trans_data.shape[0]
    else:
        num_trans_data = 0
    if not data is None:
        num_data = data.shape[0]
    else:
        num_data = 0
    with nogil:
        for i in range(num_fends):
            fend_means[i] = 0.0
        if binary == 0:
            if distance_means is None:
                for i in range(num_data):
                    fend1 = data[i, 0]
                    fend2 = data[i, 1]
                    temp = data[i, 2] / (mu * corrections[fend1] * corrections[fend2])
                    fend_means[fend1] += temp
                    fend_means[fend2] += temp
            else:
                for i in range(num_data):
                    fend1 = data[i, 0]
                    fend2 = data[i, 1]
                    temp = data[i, 2] / (distance_means[i] * corrections[fend1] * corrections[fend2])
                    fend_means[fend1] += temp
                    fend_means[fend2] += temp
            for i in range(num_trans_data):
                fend1 = trans_data[i, 0]
                fend2 = trans_data[i, 1]
                temp = trans_data[i, 2] / (trans_mu * corrections[fend1] * corrections[fend2])
                if not trans_means is None:
                    temp /= trans_means[i]
                fend_means[fend1] += temp
                fend_means[fend2] += temp
        else:
            if distance_means is None:
                for i in range(num_data):
                    fend1 = data[i, 0]
                    fend2 = data[i, 1]
                    temp = 1.0 / (mu * corrections[fend1] * corrections[fend2])
                    fend_means[fend1] += temp
                    fend_means[fend2] += temp
            else:
                for i in range(num_data):
                    fend1 = data[i, 0]
                    fend2 = data[i, 1]
                    temp = 1.0 / (distance_means[i] * corrections[fend1] * corrections[fend2])
                    fend_means[fend1] += temp
                    fend_means[fend2] += temp
            for i in range(num_trans_data):
                fend1 = trans_data[i, 0]
                fend2 = trans_data[i, 1]
                temp = 1.0 / (trans_mu * corrections[fend1] * corrections[fend2])
                if not trans_means is None:
                    temp /= trans_means[i]
                fend_means[fend1] += temp
                fend_means[fend2] += temp
    return None


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def update_express_corrections(
        np.ndarray[DTYPE_int_t, ndim=1] filt,
        np.ndarray[DTYPE_int64_t, ndim=1] interactions,
        np.ndarray[DTYPE_64_t, ndim=1] fend_means,
        np.ndarray[DTYPE_t, ndim=1] corrections,
        np.ndarray[DTYPE_64_t, ndim=1] change):
    cdef long long int i
    cdef double cost, temp
    cdef long long int num_fends = filt.shape[0]
    with nogil:
        cost = 0.0
        change[0] = 0
        for i in range(num_fends):
            if filt[i] == 0:
                continue
            temp = fend_means[i] / interactions[i]
            cost += pow(temp - 1.0, 2.0)
            temp = pow(temp, 0.5)
            if temp > 1.0:
                change[0] = max(change[0], temp - 1.0)
            else:
                change[0] = max(change[0], 1.0 - temp)
            corrections[i] *= temp
        cost = pow(cost, 0.5)
    return cost


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def calculate_v(
        np.ndarray[DTYPE_int_t, ndim=2] data,
        np.ndarray[DTYPE_int_t, ndim=2] trans_data,
        np.ndarray[DTYPE_64_t, ndim=1] counts,
        np.ndarray[DTYPE_64_t, ndim=1] trans_counts,
        np.ndarray[DTYPE_64_t, ndim=2] corrections,
        np.ndarray[DTYPE_64_t, ndim=2] v):
    cdef long long int i, fend1, fend2, num_data, num_trans_data
    cdef double correction
    if not data is None:
        num_data = data.shape[0]
    else:
        num_data = 0
    if not trans_data is None:
        num_trans = trans_data.shape[0]
    else:
        num_trans = 0
    with nogil:
        for i in range(num_data):
            fend1 = data[i, 0]
            fend2 = data[i, 1]
            correction = corrections[fend1, 0] * corrections[fend2, 0] * counts[i]
            v[fend1, 0] += correction
            v[fend2, 0] += correction
        for i in range(num_trans):
            fend1 = trans_data[i, 0]
            fend2 = trans_data[i, 1]
            correction = corrections[fend1, 0] * corrections[fend2, 0] * trans_counts[i]
            v[fend1, 0] += correction
            v[fend2, 0] += correction
    return None


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def calculate_w(
        np.ndarray[DTYPE_int_t, ndim=2] data,
        np.ndarray[DTYPE_int_t, ndim=2] trans_data,
        np.ndarray[DTYPE_64_t, ndim=1] counts,
        np.ndarray[DTYPE_64_t, ndim=1] trans_counts,
        np.ndarray[DTYPE_64_t, ndim=2] corrections,
        np.ndarray[DTYPE_64_t, ndim=2] p,
        np.ndarray[DTYPE_64_t, ndim=2] w):
    cdef long long int i, fend1, fend2, num_data, num_trans_data
    cdef double correction
    if not data is None:
        num_data = data.shape[0]
    else:
        num_data = 0
    if not trans_data is None:
        num_trans = trans_data.shape[0]
    else:
        num_trans = 0
    with nogil:
        for i in range(num_data):
            fend1 = data[i, 0]
            fend2 = data[i, 1]
            correction = corrections[fend1, 0] * corrections[fend2, 0] * counts[i]
            w[fend1, 0] += correction * p[fend2, 0]
            w[fend2, 0] += correction * p[fend1, 0]
        for i in range(num_trans):
            fend1 = trans_data[i, 0]
            fend2 = trans_data[i, 1]
            correction = corrections[fend1, 0] * corrections[fend2, 0] * trans_counts[i]
            w[fend1, 0] += correction * p[fend2, 0]
            w[fend2, 0] += correction * p[fend1, 0]
    return None



#!/usr/bin/env python

import os
import sys
import subprocess
import unittest
from operator import mul

import numpy
try:
    import pysam
except:
    pass

from hifive import hic_data
import h5py


class HiCData(unittest.TestCase):
    def setUp(self):
        self.mat_data = h5py.File('test/data/test_import_mat.hcd', 'r')
        self.raw_data = h5py.File('test/data/test_import_raw.hcd', 'r')
        self.bam_data = h5py.File('test/data/test_import_bam.hcd', 'r')
        self.bin_mat_data = h5py.File('test/data/test_import_bin_mat.hcd', 'r')
        self.bin_raw_data = h5py.File('test/data/test_import_bin_raw.hcd', 'r')
        self.bin_bam_data = h5py.File('test/data/test_import_bin_bam.hcd', 'r')
        self.matrix_data = h5py.File('test/data/test_import_matrix.hcd', 'r')
        self.fend_fname = 'test/data/test.fends'
        self.binned_fend_fname = 'test/data/test_binned.fends'
        self.mat_fname = 'test/data/test.mat'
        self.raw_fname = 'test/data/test.raw'
        self.bam_fname1 = 'test/data/test_hic_1.bam'
        self.bam_fname2 = 'test/data/test_hic_2.bam'
        self.matrix_fname = 'test/data/*.test.matrix'

    def test_hic_raw_data_creation(self):
        subprocess.call("./bin/hifive hic-data -q -R %s -i 500 %s test/data/test_temp.hcd" %
                        (self.raw_fname, self.fend_fname), shell=True)
        data = h5py.File('test/data/test_temp.hcd', 'r')
        self.compare_hdf5_dicts(self.raw_data, data, 'data')

    def test_hic_mat_data_creation(self):
        subprocess.call("./bin/hifive hic-data -q -M %s -i 500 %s test/data/test_temp.hcd" %
                        (self.mat_fname, self.fend_fname), shell=True)
        data = h5py.File('test/data/test_temp.hcd', 'r')
        self.compare_hdf5_dicts(self.mat_data, data, 'data')

    def test_hic_bam_data_creation(self):
        if 'pysam' not in sys.modules.keys():
            print >> sys.stderr, "pysam required for bam import"
            return None
        subprocess.call("./bin/hifive hic-data -q -S %s %s -i 500 %s test/data/test_temp.hcd" %
                        (self.bam_fname1, self.bam_fname2, self.fend_fname), shell=True)
        data = h5py.File('test/data/test_temp.hcd', 'r')
        self.compare_hdf5_dicts(self.bam_data, data, 'data')

    def test_hic_bin_raw_data_creation(self):
        subprocess.call("./bin/hifive hic-data -q -R %s -i 500 %s test/data/test_temp.hcd" %
                        (self.raw_fname, self.binned_fend_fname), shell=True)
        data = h5py.File('test/data/test_temp.hcd', 'r')
        self.compare_hdf5_dicts(self.bin_raw_data, data, 'data')

    def test_hic_bin_mat_data_creation(self):
        subprocess.call("./bin/hifive hic-data -q -M %s -i 500 %s test/data/test_temp.hcd" %
                        (self.mat_fname, self.binned_fend_fname), shell=True)
        data = h5py.File('test/data/test_temp.hcd', 'r')
        self.compare_hdf5_dicts(self.bin_mat_data, data, 'data')

    def test_hic_bin_bam_data_creation(self):
        if 'pysam' not in sys.modules.keys():
            print >> sys.stderr, "pysam required for bam import"
            return None
        subprocess.call("./bin/hifive hic-data -q -S %s %s -i 500 %s test/data/test_temp.hcd" %
                        (self.bam_fname1, self.bam_fname2, self.binned_fend_fname), shell=True)
        data = h5py.File('test/data/test_temp.hcd', 'r')
        self.compare_hdf5_dicts(self.bin_bam_data, data, 'data')

    def test_hic_matrix_data_creation(self):
        subprocess.call("./bin/hifive hic-data -q -X '%s' -i 500 %s test/data/test_temp.hcd" %
                        (self.matrix_fname, self.binned_fend_fname), shell=True)
        data = h5py.File('test/data/test_temp.hcd', 'r')
        self.compare_hdf5_dicts(self.matrix_data, data, 'data')

    def test_hic_mat_export(self):
        data = hic_data.HiCData('test/data/test_import_raw.hcd', 'r', silent=True)
        data.export_to_mat('test/data/test_temp.mat')
        data1 = []
        for line in open('test/data/test.mat', 'r'):
            data1.append(line)
        data2 = []
        for line in open('test/data/test_temp.mat', 'r'):
            data2.append(line)
        self.assertEqual(data1, data2,
                "generated mat file doesn't match original")

    def tearDown(self):
        subprocess.call('rm -f test/data/test_temp.hcd', shell=True)
        subprocess.call('rm -f test/data/test_temp.mat', shell=True)

    def compare_arrays(self, array1, array2, name):
        self.assertTrue(array1.shape == array2.shape,
            "%s shape doesn't match target value" % name)
        self.assertTrue(array1.dtype == array2.dtype,
            "%s dtype doesn't match target value" % name)
        if str(array1.dtype).count('S') + str(array1.dtype).count('a') > 0:
            self.assertTrue(numpy.sum(array1 == array2) == reduce(mul, array1.shape),
                "%s don't match target values." % name)
        else:
            self.assertTrue(numpy.allclose(array1, array2),
                "%s don't match target values" % name)
        return None

    def compare_hdf5_dicts(self, dict1, dict2, name):
        for key in dict1['/'].attrs.keys():
            if key == 'history':
                continue
            self.assertTrue(key in dict2['/'].attrs,
                "%s missing from %s attributes" % (key, name))
            value1 = dict1['/'].attrs[key]
            value2 = dict2['/'].attrs[key]
            if isinstance(value1, float):
                self.assertTrue(
                    numpy.allclose(numpy.array([value1], dtype=numpy.float32), numpy.array([value2], dtype=numpy.float32)),
                    "%s in %s doesn't match target value" % (key, name))
            elif isinstance(value1, int):
                self.assertTrue(
                    numpy.allclose(numpy.array([value1], dtype=numpy.int32), numpy.array([value2], dtype=numpy.int32)),
                    "%s in %s doesn't match target value" % (key, name))
            else:
                self.assertTrue(value1 == value2, "%s in %s doesn't match target value" % (key, name))
        for key in dict1.keys():
            self.assertTrue(key in dict2,
                "%s missing from heatmap arrays" % name)
            self.compare_arrays(dict1[key][...], dict2[key][...], name)
        return None


if __name__ == "__main__":
    unittest.main()


#!/usr/bin/env python

import sys

import numpy
import _hmm

class HMM:
    """A simple HMM implementation with dependency only on numpy able to handle discrete and gaussian-mixture HMMs."""

    def __init__(self, num_states=2, num_distributions=1, transitions=None, pi=None, distributions=None, seed=None):
        """Initialize the model with values some or all of the arguments and randomly initializing the remainder."""
        self.num_states = num_states
        self.num_dists = num_distributions
        # There is no error checking to ensure that the number of states/distributions match the shape of passed values
        if seed is None:
            self.rng = numpy.random.RandomState()
        else:
            self.rng = numpy.random.RandomState(seed=int(seed))
        if transitions is not None:
            self.transitions = numpy.array(transitions, dtype=numpy.float64)
        else:
            self.transitions = self.rng.rand(num_states, num_states).astype(numpy.float64)
        self.transitions /= numpy.sum(self.transitions, axis=1).reshape(-1, 1)
        if pi is not None:
            self.pi = numpy.array(pi, dtype=numpy.float64)
        else:
            self.pi = self.rng.rand(num_states).astype(numpy.float64)
        self.pi /= numpy.sum(self.pi, axis=0)
        # distributions should be passed as a nested list or numpy array with the following characteristics:
        # shape - [num_states, num_mixtures, 3]
        # The last dimension contains the guassian components weight, mean, and variance
        if distributions is None:
            self.distributions = self.rng.rand(self.num_states, self.num_dists, 3).astype(numpy.float64)
        else:
            self.distributions = numpy.array(distributions).astype(numpy.float64)
        self.distributions[:, :, 0] /= numpy.sum(self.distributions[:, :, 0], axis=1).reshape(-1, 1)

    def find_probabilities(self, observations):
        """Calculate the probability of observations for each state and, if applicable, for each mixture component."""
        self.probs = numpy.zeros((self.num_states, self.num_dists + 1, observations.shape[0]), dtype=numpy.float64)
        _hmm.find_probabilities(observations, self.distributions, self.probs)

    def find_alphas(self):
        """Calculate the probability of being in a given state at a given all previous observations."""
        # self.alphas [num_states, num_observations] is the probability of being in that state after observations up
        # to that point. For example self.alphas[1, 10] is the probability of being in state1 after the first 10
        # observations
        # self.scalars [num_observations] rescales each component so there is no variable overflow.
        self.alphas = numpy.zeros((self.num_states, self.probs.shape[2]), dtype=numpy.float64)
        self.scalars = numpy.zeros(self.probs.shape[2], dtype=numpy.float64)
        _hmm.find_alphas(self.probs, self.pi, self.transitions, self.alphas, self.scalars)

    def find_betas(self):
        """Calculate the probability of reaching a state given all of the future observations."""
        # self.betas [num_states, num_observations] is the probability of being in that state after observations past
        # that point. For example self.betas[1, 10] is the probability of being in state1 followed by observations 11
        # and higher.
        self.betas = numpy.zeros((self.num_states, self.probs.shape[2]), dtype=numpy.float64)
        _hmm.find_betas(self.probs, self.transitions, self.scalars, self.betas)

    def find_etas(self):
        """Calculate the probability of being in state i and time t-1 and transitioning to state j at time t."""
        # self.etas [num_states, num_states, num_observations - 1] is the probability of observing the transition at
        # time t-1 to time t from state i to state j.
        self.etas = numpy.zeros((self.num_states, self.num_states, self.probs.shape[2] - 1), dtype=numpy.float64)
        _hmm.find_etas(self.probs, self.transitions, self.alphas, self.betas, self.etas)

    def find_gammas(self):
        """Calculate the probability of on observation coming from state i at time t and mixture component m (if continuous)."""
        # self.gamma [num_states, num_mixtures, num_observations] is the probability of an observation arising
        # from a particular component in a particular state.
        self.gammas = numpy.zeros((self.num_states, self.num_dists, self.probs.shape[2]), dtype=numpy.float64)
        _hmm.find_gammas(self.probs, self.alphas, self.betas, self.gammas)

    def train(self, observations, threshold=1e-6, max_iterations=1000):
        """Optimize the model parameters based on a set of training sequences."""
        # Training is stopped when either the maximum number of iterations is reached or the largest change in a
        # parameter is smaller than the threshold.
        change = numpy.inf
        iteration = 0
        new_pi = numpy.copy(self.pi)
        new_transitions = numpy.copy(self.transitions)
        new_distributions = numpy.copy(self.distributions)
        while change > threshold and iteration < max_iterations:
            new_pi.fill(0.0)
            new_transitions.fill(0.0)
            new_distributions.fill(0.0)
            for h in range(len(observations)):
                self.find_probabilities(observations[h])
                self.find_alphas()
                self.find_betas()
                self.find_etas()
                self.find_gammas()
                if h == len(observations) - 1:
                    finalize = 1
                else:
                    finalize = 0
                change = _hmm.update_parameter_estimates(observations[h], self.pi, self.transitions,
                                                         self.distributions, new_pi, new_transitions,
                                                         new_distributions, self.etas, self.gammas, finalize)
            print >> sys.stderr, ("\rIteration: %03i   Change: %.5f") % (iteration, change),
            iteration += 1
        print >> sys.stderr, ("\n"),
        return

    def find_path(self, observations):
        self.find_probabilities(observations)
        scores = numpy.zeros((self.num_states, observations.shape[0]), dtype=numpy.float64)
        paths = numpy.zeros((self.num_states, observations.shape[0]), dtype=numpy.int32)
        states = numpy.zeros(observations.shape[0], dtype=numpy.int32)
        log_pi = numpy.zeros(self.pi.shape, dtype=numpy.float64) - numpy.inf
        where = numpy.where(self.pi > 0.0)
        log_pi[where] = numpy.log(self.pi[where])
        log_transitions = numpy.zeros(self.transitions.shape, dtype=numpy.float64) - numpy.inf
        where = numpy.where(self.transitions > 0.0)
        log_transitions[where] = numpy.log(self.transitions[where])
        log_probs = numpy.zeros((self.probs.shape[0], self.probs.shape[2]), dtype=numpy.float64) - numpy.inf
        where = numpy.where(self.probs[:, -1, :] > 0.0)
        log_probs[where] = numpy.log(self.probs[where[0], -1, where[1]])
        ll = _hmm.find_path(observations, log_pi, log_transitions, log_probs, scores, paths, states)
        return states, ll

    def generate_sequence(self, seq_len=100):
        """Generate a set of hidden states and emmissions from the model."""
        states = numpy.zeros(seq_len, dtype=numpy.int32)
        distributions = numpy.zeros(seq_len, dtype=numpy.int32)
        observations = numpy.zeros(seq_len, dtype=numpy.float64)
        pi_sums = numpy.zeros(self.pi.shape, dtype=numpy.float64)
        transition_sums = numpy.zeros(self.transitions.shape, dtype=numpy.float64)
        weight_sums = numpy.zeros((self.num_states, self.num_dists), dtype=numpy.float64)
        rand_nums = self.rng.rand(seq_len, 2)
        _hmm.generate_sequence(rand_nums, self.pi, self.transitions, self.distributions, pi_sums,
                               transition_sums, weight_sums, states, distributions)
        observations[:] = self.rng.normal(self.distributions[states, distributions, 1],
                                          self.distributions[states, distributions, 2] ** 0.5)
        return states, observations

def test():
    hmm1 = HMM(
        seed=2001,
        num_distributions=2,
        num_states=2,
        pi=[0.5, 0.5],
        transitions=[[0.95, 0.05], [0.1, 0.9]],
        distributions=[[[0.5, 1.5, 0.25], [0.5, -1.0, 0.25]], [[0.25, 2.5, 0.25], [0.75, 4.5, 0.5]]],
        )
    observations = []
    for i in range(100):
        temp = hmm1.generate_sequence(100)
        observations.append(temp[1])
    hmm2 = HMM(
        seed=2002,
        num_distributions=2,
        num_states=2,
        )
    hmm2.train(observations, threshold=1e-6, max_iterations=400)
    print 'True Pi: %s    Learned Pi: %s' % (str(list(hmm1.pi)), str(list(hmm2.pi)))
    for i in range(hmm1.num_states):
        print 'State %i -' % i
        print 'True transitions: %s    Learned transitions: %s' % (str(list(hmm1.transitions[i, :])),
                                                                   str(list(hmm2.transitions[i, :])))
        print 'True distribution weights: %s    Learned distribution weights %s' % (
            str(list(hmm1.distributions[i, :, 0])),
            str(list(hmm2.distributions[i, :, 0])))
        print 'True distribution means: %s    Learned distribution means %s' % (
            str(list(hmm1.distributions[i, :, 1])),
            str(list(hmm2.distributions[i, :, 1])))
        print 'True distribution variances: %s    Learned distribution variances %s' % (
            str(list(hmm1.distributions[i, :, 2])),
            str(list(hmm2.distributions[i, :, 2])))



#!/usr/bin/env python

"""A class for handling 5C read data."""

import os
import sys

import numpy
import h5py
try:
    import pysam
except:
    pass



class FiveCData(object):

    """
    This class handles interaction count data for 5C experiments.

    The FiveCData class contains all of the interaction information for a 5C experiment, including pairs of fragment indices and their associated counts.
   
    .. note::
      This class is also available as hifive.FiveCData

    When initialized, this class creates an h5dict in which to store all data associated with this object.
    
    :param filename: The file name of the h5dict. This should end with the suffix '.hdf5'
    :type filename: str.
    :param mode: The mode to open the h5dict with. This should be 'w' for creating or overwriting an h5dict with name given in filename.
    :type mode: str.
    :param silent: Indicates whether to print information about function execution for this object.
    :type silent: bool.
    :returns: :class:`FiveCData` class object.

    :Attributes: * **file** (*str.*) A string containing the name of the file passed during object creation for saving the object to.
                 * **silent** (*bool.*) - A boolean indicating whether to suppress all of the output messages.
                 * **history** (*str.*) - A string containing all of the commands executed on this object and their outcomes.
    """

    def __init__(self, filename, mode='r', silent=False):
        """Create a :class:`FiveCData` object."""
        self.file = os.path.abspath(filename)
        self.silent = silent
        self.history = ''
        self.filetype = 'fivec_data'
        if mode != 'w':
            self.load()
        return None

    def __getitem__(self, key):
        """Dictionary-like lookup."""
        if key in self.__dict__:
            return self.__dict__[key]
        else:
            return None

    def __setitem__(self, key, value):
        """Dictionary-like value setting."""
        self.__dict__[key] = value
        return None

    def save(self):
        """
        Save analysis parameters to h5dict.

        :returns: None
        """
        self.history.replace("'None'", "None")
        datafile = h5py.File(self.file, 'w')
        for key in self.__dict__.keys():
            if key in ['file', 'chr2int', 'frags', 'silent']:
                continue
            elif isinstance(self[key], numpy.ndarray):
                datafile.create_dataset(key, data=self[key])
            elif not isinstance(self[key], dict):
                datafile.attrs[key] = self[key]
        datafile.close()
        return None

    def load(self):
        """
        Load data from h5dict specified at object creation.

        Any call of this function will overwrite current object data with values from the last :func:`save` call.

        :returns: None
        """
        datafile = h5py.File(self.file, 'r')
        for key in datafile.keys():
            self[key] = numpy.copy(datafile[key])
        for key in datafile['/'].attrs.keys():
            self[key] = datafile['/'].attrs[key]
        # ensure fend h5dict exists
        if 'fragfilename' in self.__dict__:
            fragfilename = self.fragfilename
            if fragfilename[:2] == './':
                fragfilename = fragfilename[2:]
            parent_count = fragfilename.count('../')
            fragfilename = '/'.join(self.file.split('/')[:-(1 + parent_count)] +
                                fragfilename.lstrip('/').split('/')[parent_count:])
            if not os.path.exists(fragfilename):
                if not self.silent:
                    print >> sys.stderr, ("Could not find %s. No fragments loaded.\n") % (fragfilename),
            else:
                self.frags = h5py.File(fragfilename, 'r')
                # create dictionary for converting chromosome names to indices
                self.chr2int = {}
                for i, chrom in enumerate(self.frags['chromosomes']):
                    self.chr2int[chrom] = i
        datafile.close()
        return None

    def load_data_from_counts(self, fragfilename, filelist):
        """
        Read interaction counts from a text file(s) and place in h5dict.

        :param fragfilename: This specifies the file name of the :class:`Fragment` object to associate with the dataset.
        :type fragfilename: str.
        :param filelist: A list containing all of the file names of counts text files to be included in the dataset. If only one file is needed, this may be passed as a string.
        :type filelist: list
        :returns: None

        :Attributes: * **fragfilename** (*str.*) - A string containing the relative path of the fragment file.
                     * **cis_data** (*ndarray*) - A numpy array of type int32 and shape N x 3 where N is the number of valid non-zero intra-regional fragment pairings observed in the data. The first column contains the fragment index (from the 'fragments' array in the Fragment object) of the upstream fragment, the second column contains the idnex of the downstream fragment, and the third column contains the number of reads observed for that fragment pair.
                     * **cis_indices** (*ndarray*) - A numpy array of type int64 and a length of the number of fragments + 1. Each position contains the first entry for the correspondingly-indexed fragment in the first column of 'cis_data'. For example, all of the downstream cis interactions for the fragment at index 5 in the Fragment object 'fragments' array are in cis_data[cis_indices[5]:cis_indices[6], :]. 
                     * **trans_data** (*ndarray*) - A numpy array of type int32 and shape N x 3 where N is the number of valid non-zero inter-regional fragment pairings observed in the data. The first column contains the fragment index (from the 'fragments' array in the Fragment object) of the upstream fragment (upstream also refers to the lower indexed chromosome in this context), the second column contains the index of the downstream fragment, and the third column contains the number of reads observed for that fragment pair.
                     * **trans_indices** (*ndarray*) - A numpy array of type int64 and a length of the number of fragments + 1. Each position contains the first entry for the correspondingly-indexed fragment in the first column of 'trans_data'. For example, all of the downstream trans interactions for the fragment at index 5 in the Fragment object 'fragments' array are in cis_data[cis_indices[5]:cis_indices[6], :].
                     * **frags** (*ndarray*) - A filestream to the hdf5 Fragment file such that all saved Fragment attributes can be accessed through this class attribute.

        When data is loaded the 'history' attribute is updated to include the history of the Fragment file that becomes associated with it.
        """
        self.history += "FiveCData.load_data_from_counts(fragfilename='%s', filelist=%s) - " % (fragfilename, str(filelist))
        # determine if fragment file exists and if so, load it
        if not os.path.exists(fragfilename):
            if not self.silent:
                print >> sys.stderr, \
                ("The fragment file %s was not found. No data was loaded.\n") % (fragfilename),
            self.history += "Error: '%s' not found\n" % fragfilename
            return None
        self.fragfilename = "%s/%s" % (os.path.relpath(os.path.dirname(os.path.abspath(fragfilename)),
                                       os.path.dirname(self.file)), os.path.basename(fragfilename))
        self.frags = h5py.File(fragfilename, 'r')
        self.history = self.frags['/'].attrs['history'] + self.history
        strands = self.frags['fragments']['strand'][...]
        chr2int = {}
        for i, j in enumerate(self.frags['chromosomes'][:]):
            chr2int[j] = i
        # create fragment name dictionary
        names = {}
        for i in range(self.frags['fragments'].shape[0]):
            names[self.frags['fragments']['name'][i]] = i
        # load data from all files, skipping if name not in the fragment file.
        if isinstance(filelist, str):
            filelist = [filelist]
        total_reads = 0
        data = {}
        for fname in filelist:
            reads = 0
            if not os.path.exists(fname):
                if not self.silent:
                    print >> sys.stderr, ("The file %s was not found...skipped.\n") % (fname.split('/')[-1]),
                self.history += "Error: '%s' not found, " % fname
                continue
            if not self.silent:
                print >> sys.stderr, ("Loading data from %s...") % (fname.split('/')[-1]),
            input = open(fname, 'r')
            for line in input:
                temp = line.strip('\n').split('\t')
                if temp[0] not in names or temp[1] not in names or temp[0] == temp[1]:
                    continue
                frag1 = names[temp[0]]
                frag2 = names[temp[1]]
                # if both in same orientation, skip
                if strands[frag1] == strands[frag2]:
                    continue
                pair = (min(frag1, frag2), max(frag1, frag2))
                if pair not in data:
                    data[pair] = 0
                data[pair] = int(temp[2])
                reads += int(temp[2])
            input.close()
            if not self.silent:
                print >> sys.stderr, ("%i validly-mapped reads loaded.\n") % (reads),
            total_reads += reads
        if len(data) == 0:
            if not self.silent:
                print >> sys.stderr, ("No valid data was loaded.\n"),
            return None
            self.history += "Error: no valid data loaded\n"
        if not self.silent:
            print >> sys.stderr, ("%i total validly-mapped read pairs loaded. %i unique pairs\n") %\
                             (total_reads,len(data)),
        # write fragment pairs to h5dict
        self._parse_fragment_pairs(data)
        self.history += "Success\n"
        return None

    def load_data_from_bam(self, fragfilename, filelist):
        """
        Read interaction counts from pairs of BAM files and place in h5dict.

        :param fragfilename: This specifies the file name of the :class:`Fragment` object to associate with the dataset.
        :type fragfilename: str.
        :param filelist: A list containing lists of paired read end files.
        :type filelist: list
        :returns: None

        :Attributes: * **fragfilename** (*str.*) - A string containing the relative path of the fragment file.
                     * **cis_data** (*ndarray*) - A numpy array of type int32 and shape N x 3 where N is the number of valid non-zero intra-regional fragment pairings observed in the data. The first column contains the fragment index (from the 'fragments' array in the Fragment object) of the upstream fragment, the second column contains the idnex of the downstream fragment, and the third column contains the number of reads observed for that fragment pair.
                     * **cis_indices** (*ndarray*) - A numpy array of type int64 and a length of the number of fragments + 1. Each position contains the first entry for the correspondingly-indexed fragment in the first column of 'cis_data'. For example, all of the downstream cis interactions for the fragment at index 5 in the Fragment object 'fragments' array are in cis_data[cis_indices[5]:cis_indices[6], :]. 
                     * **trans_data** (*ndarray*) - A numpy array of type int32 and shape N x 3 where N is the number of valid non-zero inter-regional fragment pairings observed in the data. The first column contains the fragment index (from the 'fragments' array in the Fragment object) of the upstream fragment (upstream also refers to the lower indexed chromosome in this context), the second column contains the idnex of the downstream fragment, and the third column contains the number of reads observed for that fragment pair.
                     * **trans_indices** (*ndarray*) - A numpy array of type int64 and a length of the number of fragments + 1. Each position contains the first entry for the correspondingly-indexed fragment in the first column of 'trans_data'. For example, all of the downstream trans interactions for the fragment at index 5 in the Fragment object 'fragments' array are in cis_data[cis_indices[5]:cis_indices[6], :].
                     * **frags** (*filestream*) - A filestream to the hdf5 Fragment file such that all saved Fragment attributes can be accessed through this class attribute.

        When data is loaded the 'history' attribute is updated to include the history of the Fragment file that becomes associated with it.
        """
        self.history += "FiveCData.load_data_from_counts(fragfilename='%s', filelist=%s) - " % (fragfilename, str(filelist))
        if 'pysam' not in sys.modules.keys():
            if not self.silent:
                print >> sys.stderr, ("The pysam module must be installed to use this function.")
            self.history += 'Error: pysam module missing\n'
            return None
        # determine if fragment file exists and if so, load it
        if not os.path.exists(fragfilename):
            if not self.silent:
                print >> sys.stderr, ("The fragment file %s was not found. No data was loaded.\n") % (fragfilename),
            self.history += "Error: '%s' not found\n" % fragfilename
            return None
        self.fragfilename = "%s/%s" % (os.path.relpath(os.path.dirname(os.path.abspath(fragfilename)),
                                       os.path.dirname(self.file)), os.path.basename(fragfilename))
        self.frags = h5py.File(fragfilename, 'r')
        strands = self.frags['fragments']['strand'][...]
        chr2int = {}
        for i, j in enumerate(self.frags['chromosomes'][:]):
            chr2int[j] = i
        # create fragment name dictionary
        names = {}
        for i in range(self.frags['fragments'].shape[0]):
            names[self.frags['fragments']['name'][i]] = i
        # load data from all files, skipping if either fragment not not in the fragment file.
        if isinstance(filelist[0], str):
            filelist = [[filelist[0], filelist[1]]]
        total_reads = 0
        data = {}
        for filepair in filelist:
            # determine which files have both mapped ends present
            present = True
            if not os.path.exists(filepair[0]):
                if not self.silent:
                    print >> sys.stderr, ("%s could not be located.") % (filepair[0]),
                self.history += "'%s' not found, " % filepair[0]
                present = False
            if not os.path.exists(filepair[1]):
                if not self.silent:
                    print >> sys.stderr, ("%s could not be located.") % (filepair[1]),
                self.history += "'%s' not found, " % filepair[1]
                present = False
            if not present:
                if not self.silent:
                    print >> sys.stderr, ("No data for one or both ends could be located. Skipping this run.\n")
            reads = 0
            unpaired = {}
            # load first half of paired ends
            if not self.silent:
                print >> sys.stderr, ("Loading data from %s...") % (filepair[0].split('/')[-1]),
            input = pysam.Samfile(filepair[0], 'rb')
            for read in input.fetch(until_eof=True):
                # Only consider reads with an alignment
                if read.is_unmapped:
                    continue
                # if mapping name not in fragment names, skip
                seq_name = input.getrname(read.tid)
                if seq_name not in names:
                    continue
                # skip multiply-aligned reads
                for tag in read.tags:
                    if tag[0] == 'XS':
                        break
                else:
                    unpaired[read.qname] = names[seq_name]
            input.close()
            if not self.silent:
                print >> sys.stderr, ("Done\n"),
            # load second half of paired ends
            if not self.silent:
                print >> sys.stderr, ("Loading data from %s...") % (filepair[1].split('/')[-1]),
            input = pysam.Samfile(filepair[1], 'rb')
            for read in input.fetch(until_eof=True):
                # Only consinder reads whose paired end was valid
                if read.qname not in unpaired:
                    continue
                # Only consider reads with an alignment
                if read.is_unmapped:
                    continue
                # if mapping name not in fragment names, skip
                seq_name = input.getrname(read.tid)
                if seq_name not in names:
                    continue
                # skip multiply-aligned reads
                for tag in read.tags:
                    if tag[0] == 'XS':
                        break
                else:
                    # if both ends map to the same orientation, skip
                    if strands[unpaired[read.qname]] != strands[names[seq_name]]:
                        pair = (min(unpaired[read.qname], names[seq_name]),
                                max(unpaired[read.qname], names[seq_name]))
                        if pair not in data:
                            data[pair] = 0
                        data[pair] += 1
                        reads += 1
                    del unpaired[read.qname]
            input.close()
            if not self.silent:
                print >> sys.stderr, ("Done\n"),
            if not self.silent:
                print >> sys.stderr, ("Read %i validly_mapped read paired.\n") % (reads),
            total_reads += reads           
        if len(data) == 0:
            if not self.silent:
                print >> sys.stderr, ("No valid data was loaded.\n"),
            self.history += "Error: no valid data loaded\n"
            return None
        if not self.silent:
            print >> sys.stderr, ("%i total validly-mapped read pairs loaded. %i unique pairs\n") %\
                             (total_reads,len(data)),
        self._parse_fragment_pairs(data)
        self.history += 'Success\n'
        return None

    def _parse_fragment_pairs(self, frag_pairs):
        """Separate frag pairs into cis (within region) and trans (between region) interactions and write to h5dict with index arrays."""
        if not self.silent:
            print >> sys.stderr, ("Writing fragment pair data to file..."),
        cis = {}
        trans = {}
        for i in range(self.frags['regions'].shape[0]):
            # Find pairs from same region
            for j in range(self.frags['regions']['start_frag'][i], self.frags['regions']['stop_frag'][i] - 1):
                for k in range(j + 1, self.frags['regions']['stop_frag'][i]):
                    if (j, k) in frag_pairs:
                        cis[(j, k)] = frag_pairs[(j, k)]
            # Find pairs from different regions
            for j in range(self.frags['regions']['start_frag'][i], self.frags['regions']['stop_frag'][i]):
                for k in range(self.frags['regions']['stop_frag'][i], self.frags['fragments'].shape[0]):
                    if (j, k) in frag_pairs:
                        trans[(j, k)] = frag_pairs[(j, k)]
        # convert data into arrays
        self.cis_data = numpy.empty((len(cis), 3), dtype=numpy.int32)
        self.trans_data = numpy.empty((len(trans), 3), dtype=numpy.int32)
        keys = cis.keys()
        keys.sort()
        for i in range(len(keys)):
            self.cis_data[i, 0] = keys[i][0]
            self.cis_data[i, 1] = keys[i][1]
            self.cis_data[i, 2] = cis[keys[i]]
        keys = trans.keys()
        keys.sort()
        for i in range(len(keys)):
            self.trans_data[i, 0] = keys[i][0]
            self.trans_data[i, 1] = keys[i][1]
            self.trans_data[i, 2] = trans[keys[i]]
        # find first instance of each fragment for cis and trans data
        if self.cis_data.shape[0] > 0:
            self.cis_indices = numpy.r_[0, numpy.bincount(self.cis_data[:, 0],
                                        minlength=self.frags['fragments'].shape[0])].astype(numpy.int64)
            for i in range(1, self.cis_indices.shape[0]):
                self.cis_indices[i] += self.cis_indices[i - 1]
        if self.trans_data.shape[0] > 0:
            self.trans_indices = numpy.r_[0, numpy.bincount(self.trans_data[:, 0],
                                          minlength=self.frags['fragments'].shape[0])].astype(numpy.int64)
            for i in range(1, self.trans_indices.shape[0]):
                self.trans_indices[i] += self.trans_indices[i - 1]
        if not self.silent:
            print >> sys.stderr, ("Done\n"),
        return None



#!/usr/bin/env python

import sys
import subprocess

try:
    from mpi4py import MPI
except:
    pass
try:
    from pyx import *
except:
    pass
import h5py

from ..hic import HiC
from ..plotting import plot_hic_heatmap, plot_key


def run(args):
    if 'mpi4py' in sys.modules.keys():
        comm = MPI.COMM_WORLD
        rank = comm.Get_rank()
        num_procs = comm.Get_size()
    else:
        comm = None
        rank = 0
        num_procs = 1
    if not args.image is None and args.pdf and "pyx" not in sys.modules.keys():
        if rank == 0:
            print >> sys.stderr, ("-p/--pdf requires the package 'pyx'"),
        sys.exit(1)
    if args.chroms is None:
        chroms = []
    else:
        chroms = args.chroms.split(',')
        if len(chroms) == 1 and chroms[0] == '':
            chroms = []
    hic = HiC(args.project, 'r', silent=args.silent)
    hic.write_heatmap(args.output, binsize=args.binsize, includetrans=args.trans,
                      datatype=args.datatype, chroms=chroms, dynamically_binned=args.dynamic,
                      expansion_binsize=args.expbinsize, minobservations=args.minobs,
                      searchdistance=args.search, removefailed=args.remove, format=args.format)
    if rank > 0:
        sys.exit(0)
    if not args.image is None:
        if args.format == 'txt':
            if rank == 0:
                print >> sys.stderr, ("Plotting is only available for non-txt formats.\n"),
            return None
        kwargs = {}
        for arg in args.keywords:
            temp = arg.split("=")
            if temp[1] in ["True", "TRUE", "true"]:
                temp[1] = True
            elif temp[1] in ["False", "FALSE", "false"]:
                temp[1] = False
            elif temp[1][0] == "(":
                temp[1] = temp[1].strip('()').split(',')
                for i in range(len(temp[1])):
                    temp[1][i] = int(temp[1][i])
                temp[1] = tuple(temp[1])
            elif temp[1][0] == "[":
                temp[1] = temp[1].strip('[]').split(',')
                for i in range(len(temp[1])):
                    temp[1][i] = int(temp[1][i])
            else:
                try:
                    temp[1] = int(temp[1])
                except:
                    try:
                        temp[1] = float(temp[1])
                    except:
                        # strip off extra characters introduced by galaxy into color format
                        temp[1] = temp[1].replace('__pd__','')
            kwargs[temp[0]] = temp[1]
        if 'symmetricscaling' not in kwargs:
            if args.datatype == 'enrichment':
                kwargs['symmetricscaling'] = True
            else:
                kwargs['symmetricscaling'] = False
        img, minscore, maxscore = plot_hic_heatmap(args.output, returnscale=True, silent=args.silent,
                                                   format=args.format, **kwargs)
        if not args.pdf:
            img_format = args.image.split('.')[-1].upper()
            if img_format not in ['PNG', 'TIF', 'JPG', 'JPEG']:
                img_format = 'PNG'
            img.save(args.image, img_format)
        else:
            unit.set(defaultunit="cm")
            text.set(mode="latex")
            text.preamble(r"\usepackage{times}")
            text.preamble(r"\usepackage{sansmath}")
            text.preamble(r"\sansmath")
            text.preamble(r"\renewcommand*\familydefault{\sfdefault}")
            hm = h5py.File(args.output, 'r')
            chroms = hm['chromosomes'][...]
            sizes = [0]
            minsize = 999999999
            for chrom in chroms:
                sizes.append(hm['%s.positions' % chrom].shape[0])
                minsize = min(minsize, sizes[-1])
            if len(sizes) > 2:
                sizes[1] += 0.5
                sizes[-1] += 0.5
                if len(sizes) > 3:
                    for i in range(2, len(sizes) - 1):
                        sizes[i] += 1.0
            for i in range(1, len(sizes)):
                sizes[i] += sizes[i - 1]
            height = width = max(5.0, sizes[-1] * 0.5 / minsize)
            for i in range(len(sizes)):
                sizes[i] = sizes[i] / sizes[-1] * height
            c = canvas.canvas()
            c.insert(bitmap.bitmap(0, 0, img, width=width))
            if args.legend:
                if 'min_color' in kwargs:
                    min_color = kwargs['min_color']
                else:
                    min_color = "0000ff"
                if 'mid_color' in kwargs:
                    mid_color = kwargs['mid_color']
                else:
                    mid_color = "ffffff"
                if 'max_color' in kwargs:
                    max_color = kwargs['max_color']
                else:
                    max_color = "ff0000"
                if 'logged' in kwargs:
                    logged = kwargs['logged']
                else:
                    logged = True
                c.insert(plot_key(min_score=minscore, max_score=maxscore, height=(height * 0.05),
                                  width=width, orientation='top', num_ticks=5, min_color=min_color,
                                  mid_color=mid_color, max_color=max_color, log_display=False),
                                  [trafo.translate(0, height * 1.05)])
                if logged:
                    label = "Log2 "
                else:
                    label = ""
                if args.datatype == 'enrichment':
                    c.text(width * 0.5, height * 1.1 + 0.75, "%sEnrichment" % label,
                           [text.halign.center, text.valign.bottom, text.size(-2)])
                elif args.datatype == 'raw':
                    c.text(width * 0.5, height * 1.1 + 0.75, "%sCounts" % label,
                           [text.halign.center, text.valign.bottom, text.size(-2)])
                else:
                    c.text(width * 0.5, height * 1.1 + 0.75, "%sNormalized Counts" % label,
                           [text.halign.center, text.valign.bottom, text.size(-2)])
            if args.names:
                for i, chrom in enumerate(chroms):
                    c.text(width + 0.25, height - (sizes[i] + sizes[i + 1]) / 2, 'chr%s' % chrom,
                           [text.halign.left, text.valign.middle, text.size(-2)])
            c.writePDFfile(args.image)
            if len(args.image.split('.')) <= 1 or args.image.split('.')[-1] != 'pdf':
                subprocess.call('mv %s.pdf %s' % (args.image, args.image), shell=True)



#!/usr/bin/env python

from ..fragment import Fragment


def run(args):
    fragments = Fragment(args.output, mode='w', silent=args.silent)
    fragments.load_fragments(args.bed, genome_name=args.genome, re_name=args.re)
    fragments.save()
    return None



import sys

import connect_files
import create_fragmentset
import create_fivec_dataset
import create_fivec_project
import normalize_fivec_project
import complete_fivec_project
import create_fivec_heatmap
import combine_fivec_replicates
import get_fivec_interval
import create_fendset
import create_hic_dataset
import create_hic_project
import normalize_hic_project
import complete_hic_project
import create_hic_heatmap
import create_hic_mrheatmap
import combine_hic_replicates
import get_hic_interval

#------------------------------------------------

class Std( object ):
    def __init__( self ):
        self.__dict__ = sys.modules
    def __getattr__( self, name ):
        m = __import__(name)
        return m

std = Std()



import sys

import _fivec_binning
import _fivec_optimize
import _hic_binning
import _hic_distance
import _hic_interactions
import _hic_optimize
import _hic_domains
import _hmm
import hmm

#------------------------------------------------

class Std( object ):
    def __init__( self ):
        self.__dict__ = sys.modules
    def __getattr__( self, name ):
        m = __import__(name)
        return m

std = Std()
